#!python3
"""
Script name ``jump_to_line.py``.

Shows dialog where you can enter line number to scroll to.
"""

import blackmamba.ide.source as source
import console


def main():
    try:
        input = console.input_alert('Jump to line...', 'Empty (or invalid) value  to dismiss.')
        source.scroll_to_line(int(input))
    except ValueError:
        # Invalid input value (not int)
        pass
    except KeyboardInterrupt:
        # Cancel button
        pass


if __name__ == '__main__':
    main()



#!python3

from objc_util import ObjCClass, on_main_thread
import time
import os
import json
import requests
import console
from blackmamba.log import info, error
from blackmamba.config import get_config_value
import clipboard


_DEFAULTS_LAST_UPDATE_CHECK_KEY = 'BlackMambaLastUpdateCheck'
_ROOT_DIR = os.path.expanduser('~/Documents/site-packages-3/blackmamba')
_RELEASE_PATH = os.path.join(_ROOT_DIR, '.release.json')
_OWNER = 'zrzka'
_REPOSITORY = 'blackmamba'
_INSTALL_COMMAND = 'import requests as r; exec(r.get(\'http://bit.ly/get-blackmamba\').text)'


def _user_defaults():
    NSUserDefaults = ObjCClass('NSUserDefaults')
    return NSUserDefaults.standardUserDefaults()


@on_main_thread
def _get_last_update_check():
    return _user_defaults().integerForKey_(_DEFAULTS_LAST_UPDATE_CHECK_KEY)


@on_main_thread
def _set_last_update_check(timestamp):
    _user_defaults().setInteger_forKey_(timestamp, _DEFAULTS_LAST_UPDATE_CHECK_KEY)


def _timestamp():
    return int(time.time())


def get_local_release():
    try:
        with open(_RELEASE_PATH, 'rt') as input:
            return json.load(input)
    except:
        pass


def _url(command):
    return 'https://api.github.com/repos/{}/{}/{}'.format(_OWNER, _REPOSITORY, command)


def _request(method, command):
    url = _url(command)
    headers = {
        'Accept': 'application/vnd.github.v3+json'
    }

    return requests.request(method, url, headers=headers)


def _get(command):
    return _request('GET', command)


def _get_json(command):
    return _get(command).json()


def _get_latest_release():
    try:
        return _get_json('releases/latest')
    except:
        pass


def check():
    if not get_config_value('update.enabled', True):
        return

    timestamp = _timestamp()
#     last_check = timestamp - check_interval
    last_check = _get_last_update_check() or timestamp
    if last_check + get_config_value('update.interval', 86400) > timestamp:
        return
    _set_last_update_check(timestamp)

    info('Checking for Black Mamba updates...')

    local_release = get_local_release()
    latest_release = _get_latest_release()

    if not latest_release:
        error('Failed to fetch latest release version info')
        return

    info('Latest Black Mamba release {} (tag {})'.format(latest_release['name'], latest_release['tag_name']))

    if local_release:
        info('Installed version {} (tag {})'.format(local_release['name'], local_release['tag_name']))

        if local_release['tag_name'] == latest_release['tag_name']:
            info('No updates available, you are up to date')
            return

        clipboard.set(_INSTALL_COMMAND)

        console.alert('Black Mamba',
                      'New version {} (tag {}) available.'.format(latest_release['name'], latest_release['tag_name']),
                      'Got it!', hide_cancel_button=True)
    else:
        info('Missing installed version info, you should use the installer')



#!python3
"""
Black Mamba initialization module.

Module is Python 3 compatible only.

The only requirement is to call :func:`blackmamba.main`. Example:

.. code-block:: python

    import blackmamba
    blackmamba.main()

.. warning:: Do not add top level imports which depends on Pythonista modules. Module
    must be importable on any other platform. Add these imports to specific functions
    decorated with :obj:`blackmamba.system.Pythonista` and / or
    :obj:`blackmamba.system.iOS`.

Reference
=========
"""

from blackmamba.log import info, error, get_level, set_level, ERROR
import blackmamba.system as system

__version__ = '1.0.0'
__author__ = 'Robert Vojta'

_LATEST_VERSION_COMPATIBILITY_TEST = (311009, '3.1.1')


def _register_key_command(input, modifier, function, title=None):
    from blackmamba.uikit.keyboard import register_key_command
    import blackmamba.ide.script as script
    import os

    def make_run_script(path):
        path = os.path.join(
            os.path.expanduser('site-packages-3/blackmamba/script'),
            path
        )

        def run():
            script.run_script(path)

        return run

    if isinstance(function, str):
        function = make_run_script(function)

    register_key_command(input, modifier, function, title)


@system.Pythonista(appex=False)
@system.iOS('11.0')
def _register_ios11_default_key_commands():
    from blackmamba.uikit.keyboard import UIKeyModifier

    commands = [
        ('E', UIKeyModifier.command,
         'drag_and_drop.py', 'Drag & Drop')
    ]

    for command in commands:
        _register_key_command(*command)


@system.Pythonista(appex=False)
def _register_default_key_commands():
    from blackmamba.uikit.keyboard import UIKeyModifier, UIKeyInput
    import blackmamba.ide.tab as tab
    import blackmamba.ide.source as source

    info('Registering default key commands...')

    commands = [
        # Scripts allowed to be used in the wrench
        ('/', UIKeyModifier.command,
         'toggle_comments.py', 'Toggle Comments'),
        ('N', UIKeyModifier.command,
         'new_file.py', 'New File'),
        ('T', UIKeyModifier.command,
         'new_tab.py', 'New Tab'),
        ('W', UIKeyModifier.command | UIKeyModifier.shift,
         'close_all_tabs_except_current_one.py', 'Close Tabs Except Current One'),
        ('O', UIKeyModifier.command | UIKeyModifier.shift,
         'open_quickly.py', 'Open Quickly...'),
        ('0', UIKeyModifier.command | UIKeyModifier.shift,
         'search_dash.py', 'Search in Dash'),
        ('R', UIKeyModifier.command | UIKeyModifier.shift,
         'run_quickly.py', 'Run Quickly...'),
        ('A', UIKeyModifier.command | UIKeyModifier.shift,
         'action_quickly.py', 'Action Quickly...'),
        ('B', UIKeyModifier.control | UIKeyModifier.shift,
         'analyze.py', 'Analyze & Check Style'),
        ('K', UIKeyModifier.command | UIKeyModifier.shift,
         'clear_annotations.py', 'Clear Annotations'),
        ('U', UIKeyModifier.command,
         'run_unit_tests.py', 'Run Unit Tests...'),
        ('L', UIKeyModifier.command | UIKeyModifier.shift,
         'outline_quickly.py', 'Outline Quickly...'),
        ('L', UIKeyModifier.control,
         'jump_to_line.py', 'Jump to line...'),
        ('J', UIKeyModifier.command | UIKeyModifier.control,
         'jump_to_definition.py', 'Jump to definition...'),
        ('U', UIKeyModifier.command | UIKeyModifier.control,
         'find_usages.py', 'Find usages...'),
        ('?', UIKeyModifier.command | UIKeyModifier.control,
         'show_documentation.py', 'Show Documentation'),

        # Still keyboard only
        ('0', UIKeyModifier.command,
         tab.toggle_navigator,
         'Toggle Navigator'),
        ('W', UIKeyModifier.command,
         tab.close_selected_tab,
         'Close Tab'),
        ('\t', UIKeyModifier.control,
         tab.select_next_tab,
         'Show Next Tab'),
        ('\t', UIKeyModifier.control | UIKeyModifier.shift,
         tab.select_previous_tab,
         'Show Previous Tab'),
        (']', UIKeyModifier.command | UIKeyModifier.shift,
         tab.select_next_tab),
        ('[', UIKeyModifier.command | UIKeyModifier.shift,
         tab.select_previous_tab),
        (UIKeyInput.upArrow, UIKeyModifier.control,
         source.page_up),
        (UIKeyInput.downArrow, UIKeyModifier.control,
         source.page_down)
    ]

    for command in commands:
        _register_key_command(*command)

    def _make_select_tab(index):
        def select_tab():
            tab.select_tab(index)
        return select_tab

    for i in range(9):
        _register_key_command(str(i + 1), UIKeyModifier.command, _make_select_tab(i))

    # No need to show Cmd-[Shift]-S to users
    _log_level = get_level()
    set_level(ERROR)

    def save_all():
        tab.save(True)

    _register_key_command('S', UIKeyModifier.command, tab.save)
    _register_key_command('S', UIKeyModifier.command | UIKeyModifier.shift, save_all)
    set_level(_log_level)

    info('Default key commands registered')


@system.Pythonista(appex=False)
def _check_compatibility_and_updates():
    import blackmamba.update
    info('Pythonista {} ({})'.format(system.PYTHONISTA_VERSION, system.PYTHONISTA_BUNDLE_VERSION))

    local_release = blackmamba.update.get_local_release()
    if local_release:
        info('Black Mamba {} (tag {})'.format(__version__, local_release['tag_name']))
    else:
        info('Black Mamba {} (tag unknown, not installed via installation script)'.format(__version__))

    if system.PYTHONISTA_BUNDLE_VERSION > _LATEST_VERSION_COMPATIBILITY_TEST[0]:
        error('Installed Black Mamba version is not tested with current version of Pythonista')
        error('Latest compatibility tests were made with Pythonista {} ({})'.format(
            _LATEST_VERSION_COMPATIBILITY_TEST[1],
            _LATEST_VERSION_COMPATIBILITY_TEST[0]))
        error('Update Black Mamba or use at your own risk')

    blackmamba.update.check()


@system.Pythonista()
@system.catch_exceptions
def _main(config=None):
    # It's here because Sphinx doesn't show documentation for decorated
    # functions
    from blackmamba.config import update_config_with_dict, get_config_value
    info('Black Mamba initialization...')
    if config:
        update_config_with_dict(config)
    _check_compatibility_and_updates()
    if get_config_value('general.register_key_commands', True):
        _register_default_key_commands()
        _register_ios11_default_key_commands()
    info('Black Mamba initialized')


def main(config=None):
    """
    Black Mamba initialization.

    Call this function from ``pythonista_startup.py`` (``site-packages-3``) file.

    :param config: Optional dictionary, see :ref:`configuration`

    Example:

    .. code-block:: python

        import blackmamba

        config = {
            'general': {
                'jedi': True
            }
        }

        blackmamba.main(config)

    See `pythonista_startup.py <https://github.com/zrzka/blackmamba/blob/master/pythonista_startup.py>`_
    for more examples.

    """
    _main(config)


if __name__ == '__main__':
    main()



#!python3
"""
Script name ``search_dash.py``.

Opens `Dash <https://kapeli.com/dash_ios>`_ application with selected text or a symbol around cursor position.
"""

import editor
import urllib.parse
import webbrowser
import re

_ALLOWED_CHAR = re.compile('[a-zA-Z0-9_]')


def _encode_query(query):
    return urllib.parse.quote_plus(query, safe='', encoding=None, errors=None)


def _dash_url(query):
    return 'dash://' + _encode_query(query)


def _allowed_char(ch):
    return _ALLOWED_CHAR.match(ch) is not None


def _query():
    text = editor.get_text()

    if not text:
        return None

    selection = editor.get_selection()

    # Begin/end differs, query for selection
    if not selection[0] == selection[1]:
        return text[selection[0]:selection[1]]

    # Try to select current keyword around current cursor position
    bi = selection[0]
    while bi > 0 and _allowed_char(text[bi - 1:bi]):
        bi -= 1

    ei = selection[1]
    while ei < len(text) and _allowed_char(text[ei:ei + 1]):
        ei += 1

    return text[bi:ei]


def main(query=None):
    """Opens Dash documentation browser.

    If you do not provide query, current selection will be used. If there's
    nothing selected in the editor, cursor position is used to find identifier.

    Nothing happens if query is not provided and identifier can't be found.

    Args:
        query: Text to search in Dash
    """

    q = query or _query()

    if not q:
        return

    webbrowser.open(_dash_url(q))


if __name__ == '__main__':
    main()



"""
Following scripts are usable as Pythonista action items (wrench menu icon). In other
words, you can use them even without external keyboard.


Action quickly
==============

.. automodule:: blackmamba.script.action_quickly

Analyze
=======

.. automodule:: blackmamba.script.analyze

Clear annotations
=================

.. automodule:: blackmamba.script.clear_annotations

Close all tabs except current one
=================================

.. automodule:: blackmamba.script.close_all_tabs_except_current_one


Drag & Drop
===========

.. automodule:: blackmamba.script.drag_and_drop


Find usages
===========

.. automodule:: blackmamba.script.find_usages


Jump to definition
==================

.. automodule:: blackmamba.script.jump_to_definition


Jump to line
============

.. automodule:: blackmamba.script.jump_to_line


New file
========

.. automodule:: blackmamba.script.new_file


New tab
=======

.. automodule:: blackmamba.script.new_tab


Open quickly
============

.. automodule:: blackmamba.script.open_quickly


Outline quickly
===============

.. automodule:: blackmamba.script.outline_quickly


Run quickly
===========

.. automodule:: blackmamba.script.run_quickly


Run unit tests
==============
.. automodule:: blackmamba.script.run_unit_tests


Search Dash
===========

.. automodule:: blackmamba.script.search_dash


Show documentation
==================

.. automodule:: blackmamba.script.show_documentation


Toggle comments
===============

.. automodule:: blackmamba.script.toggle_comments
"""



#!python3

import editor
from blackmamba.config import get_config_value


def get_line_count():
    text = editor.get_text()

    if text is None:
        return None

    return len(text.splitlines())


def get_line_number():
    text = editor.get_text()

    if text is None:
        return None

    return text.count('\n', 0, editor.get_selection()[0]) + 1


def get_column_index():
    text = editor.get_text()

    if text is None:
        return None

    col = 0
    index = editor.get_selection()[0]
    while index > 0:
        if text[index - 1] == '\n':
            break
        index -= 1
        col += 1

    return col


def scroll_to_line(line_number, relative=False):
    text = editor.get_text()
    if not text:
        return

    if relative:
        current_line = get_line_number()
        line_count = get_line_count()

        if current_line is None or line_count is None:
            return

        line_number = max(min(current_line + line_number, line_count), 1)

    # https://github.com/omz/Pythonista-Issues/issues/365
    start = 0
    for index, line in enumerate(text.splitlines(True)):
        if index == line_number - 1:
            editor.set_selection(start)
            return
        start += len(line)
    editor.set_selection(start)


def page_up():
    scroll_to_line(-get_config_value('general.page_line_count', 40), True)


def page_down():
    scroll_to_line(get_config_value('general.page_line_count', 40), True)



#!python3
"""
Script name ``run_quickly.py``.

Shows dialog with all your Python files. You can filter these files by directories,
file name, etc. Use arrows keys to change selection and then hit ``Enter`` to
run selected file.

This script is configurable, see :ref:`configuration`.
"""

import os
from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
import blackmamba.ide.script as script
from blackmamba.config import get_config_value
import blackmamba.util.path as path


_IGNORE_FOLDERS = {
    '': ['.git'],
    '.': ['.Trash', 'Examples',
          'site-packages', 'site-packages-2', 'site-packages-3']
}


def _ignore_folders():
    return get_config_value('file_picker.ignore_folders', _IGNORE_FOLDERS)


class FilePickerItem(PickerItem):
    def __init__(self, folder, name, display_folder):
        super().__init__(name, display_folder)
        self._folder = folder

    @property
    def file_path(self):
        return os.path.join(self._folder, self.title)

    @property
    def match_value(self):
        return path.strip_documents_folder(self.file_path).lower()


class FilePickerDataSource(PickerDataSource):
    def __init__(self, allow_file=None, ignore_folders=None):
        super().__init__()
        self._root_folder = os.path.expanduser('~/Documents')

        def expand_folder(f):
            if not f:
                return f

            return os.path.normpath(os.path.join(self._root_folder, f))

        ignore_folders = {expand_folder(k): v for k, v in ignore_folders.items()}
        global_ignore_folders = ignore_folders.get('', [])

        home_folder = os.path.expanduser('~')
        items = []
        for root, subdirs, files in os.walk(self._root_folder, topdown=True, followlinks=True):
            if ignore_folders:
                ignore_list = global_ignore_folders[:]
                ignore_list.extend(ignore_folders.get(root, []))
                subdirs[:] = [d for d in subdirs if d not in ignore_list]
            display_folder = ' • '.join(root[len(home_folder) + 1:].split(os.sep))
            if allow_file:
                files = [f for f in files if allow_file(root, f)]
            items.extend([FilePickerItem(root, f, display_folder) for f in files])

        self.items = sorted(items)


def main():
    def allow_file(root, name):
        return path.is_python_file(name) and not name.startswith('.')

    def run_script(item, shift_enter):
        script.run_script(item.file_path, full_path=True, delay=1.0)

    kwargs = {
        'ignore_folders': _ignore_folders(),
        'allow_file': allow_file
    }

    v = PickerView()
    v.name = 'Run Quickly...'
    v.datasource = FilePickerDataSource(**kwargs)
    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - run Python script'
        '\n'
        'Esc - close • Cmd . - close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter scripts...'
    v.did_select_item_action = run_script
    v.present('sheet')
    v.wait_modal()


if __name__ == '__main__':
    main()



#!python3

import os

_DOCUMENTS = os.path.expanduser('~/Documents')


def strip_documents_folder(path):
    """
    Strip ~/Documents part of the path.

    ~/Documents/hallo.py -> hallo.py
    ~/Documents/folder/a.py -> folder/a.py
    """
    if path.startswith(_DOCUMENTS):
        return path[len(_DOCUMENTS) + 1:]
    return path


def is_python_file(path):
    _, ext = os.path.splitext(path)
    return ext.lower() == '.py'



#!python3
"""
Script name ``show_documentation.py``.

Shows documentation for the symbol around cursor. If definition can't be found,
HUD informs you. If there're more than one definitions, dialog appears where
you can select which one to show.

Documentation is displayed as Pythonista annotation.

JEDI must be enabled, see :ref:`configuration`.
"""

from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
import editor
import console
import blackmamba.ide.tab as tab
import blackmamba.ide.source as source
import os
import jedi
from blackmamba.config import get_config_value
import blackmamba.log as log


class LocationPickerItem(PickerItem):
    def __init__(self, path, line, display_folder, definition):
        super().__init__(
            '{}, line {}'.format(os.path.basename(path), line),
            display_folder
        )
        self.path = path
        self.line = line
        self.definition = definition

    def matches_title(self, terms):
        if not terms:
            return True

        start = 0
        for t in terms:
            start = self.path.find(t, start)

            if start == -1:
                return False

            start += len(t)

        return True


class LocationDataSource(PickerDataSource):
    def __init__(self, definitions):
        super().__init__()

        items = [
            LocationPickerItem(
                definition.module_name,
                definition.line,
                definition.full_name,
                definition
            )
            for definition in definitions
        ]

        self.items = items


def _show_documentation(definition):
    line = source.get_line_number()
    editor.annotate_line(line, definition.docstring(), 'success')


def _select_location(definitions):
    def action(item, shift_enter):
        _show_documentation(item.definition)

    v = PickerView()
    v.name = 'Multiple definitions found'
    v.datasource = LocationDataSource(definitions)

    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - show documentation'
        '\n'
        'Esc - close • Cmd . - close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter files...'
    v.did_select_item_action = action
    v.present('sheet')
    v.wait_modal()


def main():
    if not get_config_value('general.jedi', False):
        log.warn('show_documentation disabled, you can enable it by setting general.jedi to True')
        return

    path = editor.get_path()
    if path is None:
        return

    if not path.endswith('.py'):
        return

    tab.save()

    text = editor.get_text()
    if not text:
        return

    line = source.get_line_number()
    column = source.get_column_index()

    if line is None or column is None:
        return

    script = jedi.api.Script(text, line, column, path)
    definitions = [
        d for d in script.goto_definitions()
        if d.module_path and d.line
    ]

    if not definitions:
        console.hud_alert('Documentation not found', 'error')
        return

    if len(definitions) == 1:
        _show_documentation(definitions[0])
    else:
        _select_location(definitions)


if __name__ == '__main__':
    main()



#!python3
"""
Script name ``drag_and_drop.py``. **iOS >= 11.0 required**.

Shows dialog with opened files and folders tree. You can drag a file / folder to
any other iOS application. Works well with `Working Copy <http://workingcopyapp.com/>`_
and `Kaleidoscope <https://www.kaleidoscopeapp.com/>`_ for example.

You can drag a file / folder from any other iOS application as well. But there's one
limitation, you can drop them at the folder only.

.. note:: If you drop a folder, folder tree is not refreshed. You have to close
   dialog and open it again to see a new folder. Will be fixed.

This script is configurable, see :ref:`configuration`.
"""

import os
import ui
from objc_util import ns, ObjCClass, ObjCInstance, ObjCBlock, create_objc_class
from blackmamba.log import error
import blackmamba.util.runtime as runtime
import editor
import ctypes
import zipfile
from blackmamba.uikit.keyboard import (
    register_key_event_handler, unregister_key_event_handlers,
    UIEventKeyCode, UIKeyModifier
)
import blackmamba.ide.tab as tab
import console
import io
import shutil
from blackmamba.config import get_config_value

_TMP_DIR = os.environ.get('TMPDIR', os.environ.get('TMP'))

NSIndexPath = ObjCClass('NSIndexPath')
NSItemProvider = ObjCClass('NSItemProvider')
UIDragItem = ObjCClass('UIDragItem')
NSError = ObjCClass('NSError')
NSFileHandle = ObjCClass('NSFileHandle')
UITableViewDropProposal = ObjCClass('UITableViewDropProposal')

_path_items = None
_dragged_item_path = None
_dropped_item_destination_path = None
_dropped_item_is_folder = None
_dropped_item_name = None

#
# Drag support
#


def _type_identifier(path):
    if os.path.isdir(path):
        return 'public.folder'
    elif os.path.isfile(path):
        return 'public.data'
    return None


def _suggested_name(path):
    return os.path.basename(path)


def _load_data_imp(_cmd, _block_ptr):
    global _dragged_item_path
    handler = runtime.ObjCBlockPointer(_block_ptr,
                                       argtypes=[ctypes.c_void_p, ctypes.c_bool, ctypes.c_void_p])

    if _dragged_item_path:
        NSURL = ObjCClass('NSURL')
        url = NSURL.fileURLWithPath_isDirectory_(ns(_dragged_item_path), os.path.isdir(_dragged_item_path))

        _dragged_item_path = None
    else:
        url = None

    if not url:
        error = NSError.errorWithDomain_code_userInfo_('com.robertvojta.blackmamba', 1, None)
        handler(None, None, error)
    else:
        handler(url.ptr, False, None)


_load_data = ObjCBlock(_load_data_imp, restype=ctypes.c_void_p,
                       argtypes=[ctypes.c_void_p, ctypes.c_void_p])


def tableView_itemsForBeginningDragSession_atIndexPath_(_self, _cmd, tv_ptr, session_ptr, index_path_ptr):
    global _dragged_item_path

    if not _path_items:
        return ns([]).ptr

    section = ObjCInstance(index_path_ptr).section()
    row = ObjCInstance(index_path_ptr).row()

    if section >= 0 and section < len(_path_items) and row >= 0 and row < len(_path_items[section]):
        path = _path_items[section][row]

        type_identifier = _type_identifier(path)
        if not type_identifier:
            error('Failed to provide data, file does not exists?')
            return ns([]).ptr

        suggested_name = _suggested_name(path)

        provider = NSItemProvider.alloc().init()
        provider.registerFileRepresentationForTypeIdentifier_fileOptions_visibility_loadHandler_(
            type_identifier,
            0,
            0,
            _load_data
        )

        if not provider:
            error('Failed to create item provider.')
            return ns([]).ptr

        if suggested_name:
            provider.setSuggestedName(suggested_name)

        item = UIDragItem.alloc().initWithItemProvider_(provider)
        if not item:
            error('Failed to create drag item.')
            return ns([]).ptr

        _dragged_item_path = path
        return ns([item]).ptr

    return ns([]).ptr

#
# Drop support
#


def tableView_canHandleDropSession_(_self, _cmd, tv_ptr, session_ptr):
    session = ObjCInstance(session_ptr)

    utis = ['public.data', 'public.folder']

    if session.localDragSession():
        # Disallow in-app drag & drop
        return False

    if not session.hasItemsConformingToTypeIdentifiers_(utis):
        # No supported UTIs, disallow drag
        return False

    if session.items().count() > 1:
        # Only one item is supported
        return False

    return True


def tableView_dropSessionDidUpdate_withDestinationIndexPath_(_self, _cmd, tv_ptr, session_ptr, index_path_ptr):
    tv = ObjCInstance(tv_ptr)
    session = ObjCInstance(session_ptr)

    if not index_path_ptr:
        return UITableViewDropProposal.alloc().initWithDropOperation_intent_(0, 3).ptr

    index_path = ObjCInstance(index_path_ptr)

    if len(_path_items) == 2 and index_path.section() == 0:
        return UITableViewDropProposal.alloc().initWithDropOperation_intent_(0, 3).ptr

    # UIDropOperation
    #  - cancel = 0, forbidden = 1, copy = 2, move = 3
    #
    # UITableViewDropIntent
    #  - unspecified = 0, InsertAtDestinationIndexPath = 1
    #  - InsertIntoDestinationIndexPath = 2, Automatic = 3

    if tv.hasActiveDrag():
        if session.items.count > 1:
            return UITableViewDropProposal.alloc().initWithDropOperation_intent_(0, 3).ptr
        else:
            return UITableViewDropProposal.alloc().initWithDropOperation_intent_(2, 2).ptr
    else:
        return UITableViewDropProposal.alloc().initWithDropOperation_intent_(2, 2).ptr


def _drop_folder(data_ptr, path):
    try:
        if os.path.exists(path):
            console.alert(
                '{} exists'.format(os.path.basename(path)),
                'Do you want to replace existing {}?'.format(
                    'folder' if os.path.isdir(path) else 'file'
                ),
                'Replace'
            )

            if os.path.isfile(path):
                os.remove(path)
            else:
                shutil.rmtree(path)

        data = ObjCInstance(data_ptr)
        zip_data = io.BytesIO(ctypes.string_at(data.bytes(), data.length()))
        zf = zipfile.ZipFile(zip_data)

        corrupted_file = zf.testzip()
        if corrupted_file:
            console.hud_alert('Corrupted ZIP file', 'error')

        zf.extractall(os.path.dirname(path))

        console.hud_alert('{} dropped'.format(os.path.basename(path)))

    except KeyboardInterrupt:
        pass


def _drop_file(data_ptr, path):
    try:
        if os.path.exists(path):
            console.alert(
                '{} exists'.format(os.path.basename(path)),
                'Do you want to replace existing file?',
                'Replace'
            )
        data = ObjCInstance(data_ptr)

        if not data.writeToFile_atomically_(ns(path), True):
            console.hud_alert('Failed to write file', 'error')
            return

        console.hud_alert('{} dropped'.format(os.path.basename(path)))

    except KeyboardInterrupt:
        pass


def _load_dropped_data_imp(_cmd, data_ptr, error_ptr):
    global _dropped_item_destination_path, _dropped_item_is_folder, _dropped_item_name

    if error_ptr or not _dropped_item_destination_path or not _dropped_item_name or _dropped_item_is_folder is None:
        console.hud_alert('Drop operation failed', 'error')
        _dropped_item_destination_path = None
        _dropped_item_is_folder = None
        _dropped_item_name = None
        return

    if _dropped_item_is_folder:
        _drop_folder(
            data_ptr,
            os.path.join(str(_dropped_item_destination_path), str(_dropped_item_name))
        )
    else:
        _drop_file(
            data_ptr,
            os.path.join(str(_dropped_item_destination_path), str(_dropped_item_name))
        )

    _dropped_item_destination_path = None
    _dropped_item_is_folder = None
    _dropped_item_name = None


_load_dropped_data = ObjCBlock(_load_dropped_data_imp, restype=None,
                               argtypes=[ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p])


def tableView_performDropWithCoordinator_(_self, _cmd, tv_ptr, coordinator_ptr):
    global _dropped_item_destination_path, _dropped_item_is_folder, _dropped_item_name

    coordinator = ObjCInstance(coordinator_ptr)

    index_path = coordinator.destinationIndexPath()
    if not index_path:
        return

    session = coordinator.session()

    for item in session.items():
        provider = item.itemProvider()
        name = provider.suggestedName()

        if not name:
            continue

        folder = provider.hasItemConformingToTypeIdentifier('public.folder')

        if not folder and not provider.hasItemConformingToTypeIdentifier('public.data'):
            continue

        _dropped_item_destination_path = _path_items[index_path.section()][index_path.row()]
        _dropped_item_is_folder = folder
        _dropped_item_name = name

        provider.loadDataRepresentationForTypeIdentifier_completionHandler_(
            'public.folder' if folder else 'public.data',
            _load_dropped_data
        )

#
# UI
#


class FileNode:
    def __init__(self, path, parent=None):
        assert(parent is None or isinstance(parent, FileNode))

        self.path = os.path.normpath(path)
        self.parent = parent
        self.level = parent.level + 1 if parent else 0
        self.name = os.path.basename(path)
        self.children = []

    def add_child(self, child):
        assert(isinstance(child, FileNode))
        self.children.append(child)

    def dump(self):
        print('{} {}'.format(self.level * '  ', os.path.basename(self.path)))
        for child in self.children:
            child.dump()


def build_folder_tree(folder, parent=None, ignore=None):
    node = FileNode(folder, parent=parent)

    for file in os.listdir(folder):
        path = os.path.join(folder, file)

        if not os.path.isdir(path):
            continue

        if ignore and ignore(folder, file):
            continue

        child_node = build_folder_tree(path, node, ignore)
        node.add_child(child_node)

    return node


def ignore(folder, file):
    if file.startswith('.'):
        return True

    ignore_folders = get_config_value('drag_and_drop.ignore_folders', None)
    if not ignore_folders:
        return False

    for parent, folders in ignore_folders.items():
        if not parent and file in folders:
            return True

        if parent == '.' and folder == os.path.expanduser('~/Documents') and file in folders:
            return True

        if parent == os.path.basename(folder) and file in folders:
            return True

    return False


class FolderPickerDataSource:
    def __init__(self, root_node, expanded_folder=None, files=None):
        assert(isinstance(root_node, FileNode))
        assert(root_node.level == 0)

        self.tableview = None
        self._root_node = root_node
        self._items = None

        self._files = [
            {
                'title': os.path.basename(file),
                'level': 0,
                'path': file
            }
            for file in files or []
        ]

        if self._files:
            self._file_section = 0
            self._folder_section = 1
        else:
            self._file_section = -1
            self._folder_section = 0

        self._expanded_node_paths = set((root_node.path,))
        if expanded_folder:
            sentinel = os.path.expanduser('~/Documents')

            if not expanded_folder.startswith(sentinel):
                raise ValueError('expanded_folder must start with ~/Documents')

            path = expanded_folder
            while not path == sentinel:
                self._expanded_node_paths.add(path)
                path = os.path.dirname(path)

        self._update_path_items()

    def _update_path_items(self):
        global _path_items

        _file_drag_items = [f['path'] for f in self._files]
        _folder_drag_items = [f['path'] for f in self.items]

        if _file_drag_items:
            _path_items = [_file_drag_items, _folder_drag_items]
        else:
            _path_items = [_folder_drag_items]

    def _node_dict(self, node):
        return {
            'title': os.path.basename(node.path),
            'level': node.level,
            'path': node.path,
            'node': node
        }

    def _generate_node_children_items(self, node):
        items = []

        if not node.children:
            return items

        if node.path not in self._expanded_node_paths:
            return items

        for child in node.children:
            items.append(self._node_dict(child))
            items.extend(self._generate_node_children_items(child))

        return items

    def _generate_node_items(self, node):
        items = [self._node_dict(node)]
        items.extend(self._generate_node_children_items(node))
        return items

    @property
    def items(self):
        if self._items is not None:
            return self._items

        self._items = self._generate_node_items(self._root_node)
        return self._items

    def tableview_title_for_header(self, tv, section):
        if section == self._file_section:
            return 'Editor file (drag only)'
        return 'Folders (drag & drop)'

    def tableview_number_of_sections(self, tv):
        self.tableview = tv
        return self._folder_section + 1

    def tableview_number_of_rows(self, tv, section):
        if section == self._file_section:
            return len(self._files)
        else:
            return len(self.items)

    def tableview_can_delete(self, tv, section, row):
        return False

    def tableview_can_move(self, tv, section, row):
        return False

    def tableview_did_select(self, tv, section, row):
        tv_objc = ObjCInstance(tv)
        index_path = tv_objc.indexPathForSelectedRow()

        if section == self._file_section:
            tv_objc.deselectRowAtIndexPath_animated_(index_path, True)
            return

        item = self.items[row]
        node = item.get('node', None)

        if node.path == self._root_node.path or not node.children:
            tv_objc.deselectRowAtIndexPath_animated_(index_path, True)
            return

        tv_objc.beginUpdates()
        start_index = row + 1
        try:
            self._expanded_node_paths.remove(node.path)

            index_paths = []
            for index, item in enumerate(self._items[start_index:]):
                if item['node'].path.startswith(node.path):
                    index_paths.append(NSIndexPath.indexPathForRow_inSection_(start_index + index, section))
                else:
                    break

            if index_paths:
                self._items[start_index:start_index + len(index_paths)] = []
                tv_objc.deleteRowsAtIndexPaths_withRowAnimation_(ns(index_paths), 3)

        except KeyError:
            self._expanded_node_paths.add(node.path)
            child_items = self._generate_node_children_items(node)

            index_paths = []
            for i in range(start_index, start_index + len(child_items)):
                index_paths.append(NSIndexPath.indexPathForRow_inSection_(i, section))

            self._items[start_index:start_index] = child_items
            tv_objc.insertRowsAtIndexPaths_withRowAnimation_(ns(index_paths), 3)

        tv_objc.reloadRowsAtIndexPaths_withRowAnimation_(ns([index_path]), 5)
        tv_objc.endUpdates()
        self._update_path_items()

    def tableview_cell_for_row(self, tv, section, row):
        if section == self._folder_section:
            item = self.items[row]
            node = item['node']
        else:
            item = self._files[row]
            node = None

        cell = ui.TableViewCell()

        cvb = cell.content_view.bounds

        x = 15 + cvb.x + item['level'] * 15

        if node and node.children:
            image_view = ui.ImageView()
            image_view.frame = (x, 10, 24, 24)
            image_view.image = ui.Image.named(
                'iob:arrow_down_b_24' if node.path in self._expanded_node_paths else 'iob:arrow_right_b_24'
            )
            cell.content_view.add_subview(image_view)

        x += 24 + 8

        image_view = ui.ImageView()
        image_view.frame = (x, 10, 24, 24)
        image_view.image = ui.Image.named('iob:folder_24' if node else 'iob:document_24')
        cell.content_view.add_subview(image_view)

        x += 24 + 8

        title_label = ui.Label(flex='W')
        title_label.text = item['title']
        title_label.size_to_fit()
        title_label.frame = (
            x, cvb.y + (cvb.height - title_label.height) / 2.0,
            cvb.width - (x - cvb.x) - 8, title_label.height
        )
        cell.content_view.add_subview(title_label)

        separator = ui.View(flex='W')
        separator.background_color = (0, 0, 0, 0.05)
        x = title_label.frame.x - 12 - 8
        separator.frame = (
            x, cvb.y + cvb.height - 1,
            cvb.width - (x - cvb.x), 1
        )
        cell.content_view.add_subview(separator)

        cell_objc = ObjCInstance(cell)
        cell_objc.setSelectionStyle(0)

        return cell


class DragAndDropView(ui.View):
    def __init__(self):
        self.name = 'Drag & Drop'

        self.width = min(ui.get_window_size()[0] * 0.8, 700)
        self.height = ui.get_window_size()[1] * 0.8

        path = editor.get_path()
        if path:
            expanded_folder = os.path.dirname(path)
            files = tab.get_paths()
        else:
            expanded_folder = None
            files = None

        root_node = build_folder_tree(os.path.expanduser('~/Documents'), ignore=ignore)
        data_source = FolderPickerDataSource(root_node, expanded_folder, files)

        tv = ui.TableView(frame=self.bounds, flex='WH')
        tv.delegate = data_source
        tv.data_source = data_source
        tv.allows_multiple_selection = False
        tv.allows_selection = True
        tv.allows_multiple_selection_during_editing = False
        tv.allows_selection_during_editing = False
        tv.separator_color = 'clear'
        self.add_subview(tv)

        methods = [tableView_itemsForBeginningDragSession_atIndexPath_]
        protocols = ['UITableViewDragDelegate']
        DragDelegate = create_objc_class('DragDelegate', methods=methods, protocols=protocols)
        self._drag_delegate = DragDelegate.alloc().init()

        methods = [
            tableView_canHandleDropSession_,
            tableView_dropSessionDidUpdate_withDestinationIndexPath_,
            tableView_performDropWithCoordinator_
        ]
        protocols = ['UITableViewDropDelegate']
        DropDelegate = create_objc_class('DropDelegate', methods=methods, protocols=protocols)
        self._drop_delegate = DropDelegate.alloc().init()

        tv_objc = ObjCInstance(tv)
        tv_objc.setDragDelegate_(self._drag_delegate)
        tv_objc.setDropDelegate_(self._drop_delegate)

        def handle_escape():
            self.close()

        self._handlers = [
            register_key_event_handler(UIEventKeyCode.escape, handle_escape),
            register_key_event_handler(UIEventKeyCode.dot, handle_escape,
                                       modifier=UIKeyModifier.command)
        ]

    def will_close(self):
        if self._handlers:
            unregister_key_event_handlers(self._handlers)


def main():
    tab.save(True)
    view = DragAndDropView()
    view.present('sheet')


if __name__ == '__main__':
    main()



#!python3

from objc_util import on_main_thread, ObjCClass, sel


_PASlidingContainerViewController = ObjCClass('PASlidingContainerViewController')
_PA2UniversalTextEditorViewController = ObjCClass('PA2UniversalTextEditorViewController')
_UIApplication = ObjCClass('UIApplication')


def _get_root_view_controller():
    key_window = _UIApplication.sharedApplication().keyWindow()
    return key_window.rootViewController()


def _get_tabs_view_controller():
    root = _get_root_view_controller()

    if not root.isKindOfClass_(_PASlidingContainerViewController):
        return None

    return root.detailViewController()


def _get_selected_tab_view_controller():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    tab = tabs.selectedTabViewController()
    if not tab or not tab.isKindOfClass_(_PA2UniversalTextEditorViewController):
        return None

    return tab


def _get_tab_bar_view():
    tabs = _get_tabs_view_controller()
    return tabs.tabBarView() if tabs else None


def _get_tabs_count():
    tab_bar = _get_tab_bar_view()
    return len(tab_bar.tabViews()) if tab_bar else 0


def _get_selected_tab_index():
    tab_bar = _get_tab_bar_view()
    return tab_bar.selectedTabIndex() if tab_bar else -1


@on_main_thread
def select_tab(index):
    count = _get_tabs_count()

    if index < 0 or index >= count:
        return

    tab_bar = _get_tab_bar_view()
    tab_bar_views = tab_bar.tabViews()
    tab_bar.tabSelected_(tab_bar_views[index])

    tabs = _get_tabs_view_controller()
    tabs.switchToTabViewController_(tabs.tabViewControllers()[index])


@on_main_thread
def select_next_tab():
    index = _get_selected_tab_index()

    if index == -1:
        return

    select_tab((index + 1) % _get_tabs_count())


@on_main_thread
def select_previous_tab():
    index = _get_selected_tab_index()

    if index == -1:
        return

    select_tab((index - 1) % _get_tabs_count())


@on_main_thread
def save(all=False):
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    if all:
        tabs_to_save = tabs.tabViewControllers()
    else:
        tabs_to_save = [tabs.tabViewControllers()[_get_selected_tab_index()]]

    for tab in tabs_to_save:
        if tab.respondsToSelector(sel('saveData')):
            tab.saveData()


@on_main_thread
def close_tabs_except_current():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    tabs.closeAllTabsExceptCurrent()


@on_main_thread
def close_selected_tab():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    if len(tabs.tabViewControllers()) > 1:
        tabs.closeSelectedTab_(tabs.closeSelectedTabButtonItem().ptr)


@on_main_thread
def toggle_navigator():
    root = _get_root_view_controller()
    if not root:
        return

    if root.masterVisible():
        root.hideMasterWithAnimationDuration_(0.3)
    else:
        root.showMasterWithAnimationDuration_(0.3)


@on_main_thread
def new_tab():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    tabs.addTab_(tabs.addTabButtonItem())


@on_main_thread
def new_file():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return

    tabs.addTab_(tabs.addTabButtonItem())
    tab = tabs.tabViewControllers()[-1]
    tab.addNewFile_(tab.addNewFileButton())


@on_main_thread
def get_path():
    tab = _get_selected_tab_view_controller()

    if not tab:
        return None

    return str(tab.filePath())


@on_main_thread
def get_paths():
    tabs = _get_tabs_view_controller()

    if not tabs:
        return None

    return [str(tab.filePath()) for tab in tabs.tabViewControllers()]


def open_file(path, new_tab=True):
    index = None
    paths = get_paths()

    if paths:
        try:
            index = paths.index(path)
        except ValueError:
            pass

    if index is None:
        import editor
        editor.open_file(path, new_tab=new_tab)
    else:
        select_tab(index)



#!python3
"""
System info and decorators.

.. warning:: This module must not introduce dependency on any other Black Mamba
    modules and must be importable on any other platform as well.

Reference
=========
"""

import sys
import traceback
import functools

try:
    import console
except ImportError:
    console = None


# 3.1, 301016
# 3.1.1 beta, 311008

PYTHONISTA = sys.platform == 'ios'
"""bool: ``True`` if we're running within Pythonista or ``False``."""

PYTHONISTA_VERSION = None
"""str: Pythonista version or ``None`` if we're not within Pythonista."""

PYTHONISTA_BUNDLE_VERSION = None
"""int: Pythonista bundle version or ``None`` if we're not within Pythonista."""

PYTHONISTA_VERSION_TUPLE = None
"""tuple(int): Pythonista version tuple (3, 1, 1) or ``None`` if we're not within Pythonista."""

IOS = sys.platform == 'ios'
"""bool: ``True`` if we're running within iOS or ``False``."""

IOS_VERSION = None
"""str: iOS version or ``None`` if we're not within iOS."""

IOS_VERSION_TUPLE = None
"""tuple(int): iOS version tuple (11, 0) or ``None`` if we're not within iOS."""


def _version_tuple(version):
    if not version:
        return None
    return tuple(map(int, (version.split('.'))))


if PYTHONISTA:
    import plistlib
    import os

    try:
        plist_path = os.path.abspath(os.path.join(sys.executable, '..', 'Info.plist'))
        plist = plistlib.readPlist(plist_path)

        PYTHONISTA_VERSION = plist['CFBundleShortVersionString']
        PYTHONISTA_BUNDLE_VERSION = int(plist['CFBundleVersion'])
        PYTHONISTA_VERSION_TUPLE = _version_tuple(PYTHONISTA_VERSION)
    except Exception:
        pass


if IOS:
    try:
        from objc_util import ObjCClass
        IOS_VERSION = str(ObjCClass('UIDevice').currentDevice().systemVersion())
        IOS_VERSION_TUPLE = _version_tuple(IOS_VERSION)
    except Exception:
        pass


class _Available:
    def __init__(self, from_version=None, to_version=None):
        if from_version and to_version:
            raise ValueError('Either from_version or to_version can be provided, not both')

        self._from_version = _version_tuple(from_version)
        self._to_version = _version_tuple(to_version)

    def version(self):
        raise Exception('Not implemented, return version as tuple(int)')

    def _available(self):
        current_version = self.version()
        if not current_version:
            return False

        if self._to_version:
            return current_version <= self._to_version

        if self._from_version:
            return current_version >= self._from_version

        return True

    def __call__(self, fn, *args, **kwargs):
        def func(*args, **kwargs):
            if self._available():
                return fn(*args, **kwargs)
            return None
        return func


class iOS(_Available):
    """
    Decorator to execute function under specific iOS versions.

    Return value is return value of decorated function or ``None``
    if iOS condition isn't met.

    Examples:

        Run function only within any iOS version::

            @iOS()
            def run_me():
                pass

        Run function only within iOS >= 11.0::

            @iOS('11.0')  # or @iOS(from_version='11.0')
            def run_me():
                pass

        Run function only within iOS <= 11.0::

            @iOS(None, '11.0')  # or @iOS(to_version='11.0')
            def run_me():
                pass
    """
    def version(self):
        return IOS_VERSION_TUPLE


class Pythonista(_Available):
    """
    Decorator to execute function under specific Pythonista versions.

    By default, function is not executed under application extension.
    You have to pass ``appex=True`` if you'd like to run some function
    under appex as well.

    Return value is return value of decorated function or ``None``
    if Pythonista condition isn't met.

    Examples:

        Run function only within any Pythonista version::

            @Pythonista()
            def run_me():
                pass

        Run function only within any Pythonista version and allow appex::

            @Pythonista(appex=True)
            def run_me():
                pass

        Run function only within any Pythonista version and disallow appex::

            @Pythonista(appex=False)
            def run_me():
                pass

        Run function only within Pythonista >= 3.1.1::

            @Pythonista('3.1.1')  # or @Pythonista(from_version='3.1.1')
            def run_me():
                pass

        Run function only within Pythonista <= 3.2::

            @Pythonista(None, '3.2')  # or @Pythonista(to_version='3.2')
            def run_me():
                pass
    """
    def __init__(self, from_version=None, to_version=None, appex=None):
        super().__init__(from_version, to_version)
        self._appex = appex

    def _available(self):
        available = super()._available()

        if available and self._appex is not None:
            import appex
            available = appex.is_running_extension() == self._appex

        return available

    def version(self):
        return PYTHONISTA_VERSION_TUPLE


def catch_exceptions(func):
    """
    Decorator catching all ``Exception`` exceptions and writing info to console.

    Use this decorator for functions handling keyboard shortcuts,
    keyboard events, ... to avoid Pythonista crash.

    :param func: Function to decorate
    :return: Return value of ``func``
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            if console:
                console.set_color(1, 0, 0)
            print(traceback.format_exc())
            print('Please, file an issue at {}'.format('https://github.com/zrzka/blackmamba/issues'))
            if console:
                console.set_color()

    return new_func



#!python3

from enum import Enum


class Style(str, Enum):
    error = 'error'
    warning = 'warning'


class Annotation(object):
    def __init__(self, line, text, style, filename=None):
        self.line = line
        self.text = text
        self.style = style
        self.filename = filename



#!python3
"""
Script name ``toggle_comments.py``.

Toggles current line / selected lines comments.
"""

import re
from blackmamba.system import Pythonista


def _comment_line(line, hash_prefix=''):
    stripped = line.strip()

    if stripped.startswith('#'):
        return line

    if not stripped:
        return hash_prefix + '# \n'

    return hash_prefix + '# ' + line[len(hash_prefix):]


_UNCOMMENT_RE = re.compile('\A(\s*)#( ?)(.*)\Z', re.DOTALL)


def _uncomment_line(line):
    if line.find('#') == -1:
        return line

    match = _UNCOMMENT_RE.search(line)
    if match:
        result = match.group(1) + match.group(3)
    else:
        result = line

    if not result.strip():
        result = '\n'

    return result


_HASH_INDEX_RE = re.compile('\A(\s*)')


def _hash_prefix(lines):
    prefix = None

    for line in lines:
        if not line.strip():
            continue

        match = _HASH_INDEX_RE.search(line)

        if not match:
            continue

        if prefix is None or len(match.group(1)) < len(prefix):
            prefix = match.group(1)

    if prefix is None:
        prefix = ''

    return prefix


def _toggle_lines(lines):
    if not lines:
        return lines

    if lines[0].strip().startswith('#'):
        comment = False
        hash_prefix = ''
    else:
        comment = True
        hash_prefix = _hash_prefix(lines)

    replacement = []

    for line in lines:
        if comment:
            replacement.append(_comment_line(line, hash_prefix))
        else:
            replacement.append(_uncomment_line(line))

    return replacement


@Pythonista()
def main():
    import editor

    selection_range = editor.get_selection()

    if not selection_range:
        # No file opened in the editor
        return

    text = editor.get_text()

    selected_lines_range = editor.get_line_selection()
    selected_lines_text = text[selected_lines_range[0]:selected_lines_range[1]]
    selected_lines = selected_lines_text.splitlines(True)

    last_line_deleted = False
    if len(selected_lines) > 1:
        # Ignore the last line selection if there's just cursor at the beginning of
        # this line and nothing is selected
        last_line = selected_lines[-1]

        if selected_lines_range[1] - len(last_line) == selection_range[1]:
            last_line_deleted = True
            del selected_lines[-1]
            selected_lines_range = (selected_lines_range[0], selected_lines_range[1] - len(last_line) - 1)

    replacement = ''.join(_toggle_lines(selected_lines))

    if last_line_deleted:
        replacement = replacement[:-1]

    editor.replace_text(selected_lines_range[0], selected_lines_range[1], replacement)
    editor.set_selection(selected_lines_range[0], selected_lines_range[0] + len(replacement))


if __name__ == '__main__':
    main()



#!python3
"""
Script name ``find_usages.py``.

Finds usages of a symbol. If there're no usage, HUD informs you.
Otherwise dialog appears where you can select location and scroll to it.

JEDI must be enabled, see :ref:`configuration`.
"""

from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
import editor
import console
import os
import ui
import jedi
from blackmamba.config import get_config_value
import blackmamba.log as log
import blackmamba.ide.tab as tab
import blackmamba.ide.source as source


def _open_and_scroll(path, line):
    if editor.get_path() == path:
        source.scroll_to_line(line)
    else:
        def scroll():
            source.scroll_to_line(line)
        editor.open_file(path, new_tab=True)
        ui.delay(scroll, 0.2)


class LocationPickerItem(PickerItem):
    def __init__(self, path, line, display_folder):
        super().__init__(
            '{}, line {}'.format(os.path.basename(path), line),
            display_folder
        )
        self.path = path
        self.line = line

    def matches_title(self, terms):
        if not terms:
            return True

        start = 0
        for t in terms:
            start = self.path.find(t, start)

            if start == -1:
                return False

            start += len(t)

        return True


class LocationDataSource(PickerDataSource):
    def __init__(self, definitions):
        super().__init__()

        home_folder = os.path.expanduser('~/Documents')

        items = [
            LocationPickerItem(
                definition.module_path,
                definition.line,
                ' • '.join(os.path.dirname(definition.module_path)[len(home_folder) + 1:].split(os.sep))
            )
            for definition in definitions
        ]

        self.items = items


def _select_location(definitions):
    def open_location(item, shift_enter):
        _open_and_scroll(item.path, item.line)

    v = PickerView()
    v.name = 'Usages'
    v.datasource = LocationDataSource(definitions)

    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - open file and scroll to location'
        '\n'
        'Esc - close • Cmd . - close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter files...'
    v.did_select_item_action = open_location
    v.present('sheet')
    v.wait_modal()


def main():
    if not get_config_value('general.jedi', False):
        log.warn('find_usages disabled, you can enable it by setting general.jedi to True')
        return

    path = editor.get_path()
    if path is None:
        return

    if not path.endswith('.py'):
        return

    tab.save()

    text = editor.get_text()
    if not text:
        return

    line = source.get_line_number()
    column = source.get_column_index()

    if line is None or column is None:
        return

    script = jedi.api.Script(text, line, column, path)
    definitions = [
        d for d in script.usages()
        if d.module_path and d.line
    ]

    if not definitions:
        console.hud_alert('Definition not found', 'error')
        return

    _select_location(definitions)


if __name__ == '__main__':
    main()



#!python3
"""
Script name ``jump_to_definition.py``.

Jumps to symbol definition. If definition can't be found, HUD informs you.
Otherwise it jumps to definition location or shows dialog where you can choose
location if more than one definition was found.

JEDI must be enabled, see :ref:`configuration`.
"""

from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
import editor
import console
import os
import ui
import jedi
from blackmamba.config import get_config_value
import blackmamba.log as log
import blackmamba.ide.source as source
import blackmamba.ide.tab as tab


def _open_and_scroll(path, line):
    if editor.get_path() == path:
        source.scroll_to_line(line)
    else:
        def scroll():
            source.scroll_to_line(line)
        tab.open_file(path, new_tab=True)
        ui.delay(scroll, 0.2)


class LocationPickerItem(PickerItem):
    def __init__(self, path, line, display_folder, definition):
        super().__init__(
            '{}, line {}'.format(os.path.basename(path), line),
            display_folder
        )
        self.path = path
        self.line = line
        self.definition = definition

    def matches_title(self, terms):
        if not terms:
            return True

        start = 0
        for t in terms:
            start = self.path.find(t, start)

            if start == -1:
                return False

            start += len(t)

        return True


class LocationDataSource(PickerDataSource):
    def __init__(self, definitions):
        super().__init__()

        items = [
            LocationPickerItem(
                definition.module_name,
                definition.line,
                definition.full_name,
                definition
            )
            for definition in definitions
        ]

        self.items = items


def _select_location(definitions):
    def open_location(item, shift_enter):
        definition = item.definition
        _open_and_scroll(definition.module_path, definition.line)

    v = PickerView()
    v.name = 'Multiple definitions found'
    v.datasource = LocationDataSource(definitions)

    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - open file and scroll to location'
        '\n'
        'Esc - close • Cmd . - close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter files...'
    v.did_select_item_action = open_location
    v.present('sheet')
    v.wait_modal()


def main():
    if not get_config_value('general.jedi', False):
        log.warn('jump_to_definition disabled, you can enable it by setting general.jedi to True')
        return

    path = editor.get_path()
    if path is None:
        return

    if not path.endswith('.py'):
        return

    tab.save()

    text = editor.get_text()
    if not text:
        return

    line = source.get_line_number()
    column = source.get_column_index()

    if line is None or column is None:
        return

    script = jedi.api.Script(text, line, column, path)
    definitions = [
        d for d in script.goto_definitions()
        if d.module_path and d.line
    ]

    if not definitions:
        console.hud_alert('Definition not found', 'error')
        return

    if len(definitions) == 1:
        _open_and_scroll(definitions[0].module_path, definitions[0].line)
    else:
        _select_location(definitions)


if __name__ == '__main__':
    main()



#!python3

import ui
from objc_util import ObjCInstance, ObjCClass, on_main_thread
from enum import Enum
from functools import partial


_LayoutConstraint = ObjCClass('NSLayoutConstraint')


class LayoutRelation(int, Enum):
    lessThanOrEqual = -1
    equal = 0
    greaterThanOrEqual = 1


class LayoutAttribute(int, Enum):
    notAnAttribute = 0
    left = 1
    right = 2
    top = 3
    bottom = 4
    leading = 5
    trailing = 6
    width = 7
    height = 8
    centerX = 9
    centerY = 10
    baseline = 11
    lastBaseline = 12
    firstBaseline = 13
    leftMargin = 14
    rightMargin = 15
    topMargin = 16
    bottomMargin = 17
    leadingMargin = 18
    trailingMargin = 19
    centerXWithinMargins = 20
    centerYWithinMargins = 21


class LayoutConstraintOrientation(int, Enum):
    horizontal = 0
    vertical = 1


class LayoutPriority(float, Enum):
    required = 1000
    defaultHight = 750
    defaultLow = 250
    fittingSizeLevel = 50


class _LayoutBaseAttribute:
    _relations = {
        "min": LayoutRelation.greaterThanOrEqual,
        "max": LayoutRelation.lessThanOrEqual,
        "equal": LayoutRelation.equal
    }

    def __init__(self, view, attribute, other=None, other_attribute=LayoutAttribute.notAnAttribute):
        assert(isinstance(view, LayoutProxy))
        assert(isinstance(attribute, LayoutAttribute))

        self._view = view
        self._attribute = attribute
        self._constraints = {}

        if other:
            assert(isinstance(other, ui.View))
            assert(isinstance(other_attribute, LayoutAttribute))

            self._other = other
            self._other_attribute = other_attribute
        else:
            self._other = None
            self._other_attribute = LayoutAttribute.notAnAttribute

    @property
    def view(self):
        return self._view

    @property
    def attribute(self):
        return self._attribute

    @property
    def other(self):
        return self._other

    @property
    def other_attribute(self):
        return self._other_attribute

    @property
    def superview(self):
        return self._view.superview

    @on_main_thread
    def remove_constraint(self, constraint):
        ObjCInstance(self.superview).removeConstraint_(constraint)

    @on_main_thread
    def add_constraint(self, constraint):
        ObjCInstance(self.superview).addConstraint_(constraint)

    def constraint(relation, value, priority):
        raise NotImplementedError

    def __setattr__(self, name, value):
        if name in self._relations.keys():
            constraint = self._constraints.get(name, None)
            if constraint:
                self.remove_constraint(constraint)

            if value is None:
                return

            if isinstance(value, tuple):
                priority = value[1]
                value = value[0]
            else:
                priority = LayoutPriority.required

            constraint = self.constraint(self._relations[name], value, priority)
            self._constraints[name] = constraint
            self.add_constraint(constraint)
        else:
            super().__setattr__(name, value)


class _LayoutConstantAttribute(_LayoutBaseAttribute):
    def __init__(self, view, attribute, other=None, other_attribute=LayoutAttribute.notAnAttribute):
        super().__init__(view, attribute, other, other_attribute)

    def constraint(self, relation, value, priority):
        constraint = _LayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
            self.view, int(self.attribute), int(relation), self.other, int(self.other_attribute), 1.0, value
        )
        constraint.setPriority_(priority)
        return constraint


class _LayoutMultiplierAttribute(_LayoutBaseAttribute):
    def __init__(self, view, attribute, other=None, other_attribute=LayoutAttribute.notAnAttribute):
        super().__init__(view, attribute, other, other_attribute)

    def constraint(self, relation, value, priority):
        constraint = _LayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
            self.view, int(self.attribute), int(relation), self.other, int(self.other_attribute), value, 0
        )
        constraint.setPriority_(priority)
        return constraint


class Layout:
    _definitions = {
        "width": (_LayoutConstantAttribute, LayoutAttribute.width, LayoutAttribute.notAnAttribute),
        "height": (_LayoutConstantAttribute, LayoutAttribute.height, LayoutAttribute.notAnAttribute),

        "align_center_x_to": (_LayoutConstantAttribute, LayoutAttribute.centerX, LayoutAttribute.centerX),
        "align_center_y_to": (_LayoutConstantAttribute, LayoutAttribute.centerY, LayoutAttribute.centerY),
        "align_leading_to": (_LayoutConstantAttribute, LayoutAttribute.leading, LayoutAttribute.leading),
        "align_trailing_to": (_LayoutConstantAttribute, LayoutAttribute.trailing, LayoutAttribute.trailing),
        "align_top_to": (_LayoutConstantAttribute, LayoutAttribute.top, LayoutAttribute.top),
        "align_bottom_to": (_LayoutConstantAttribute, LayoutAttribute.bottom, LayoutAttribute.bottom),
        "align_left_to": (_LayoutConstantAttribute, LayoutAttribute.left, LayoutAttribute.left),
        "align_right_to": (_LayoutConstantAttribute, LayoutAttribute.right, LayoutAttribute.right),
        "align_baseline_to": (_LayoutConstantAttribute, LayoutAttribute.baseline, LayoutAttribute.baseline),

        "align_center_x_with_superview": (_LayoutConstantAttribute, LayoutAttribute.centerX, LayoutAttribute.centerX),
        "align_center_y_with_superview": (_LayoutConstantAttribute, LayoutAttribute.centerY, LayoutAttribute.centerY),
        "align_leading_with_superview": (_LayoutConstantAttribute, LayoutAttribute.leading, LayoutAttribute.leading),
        "align_trailing_with_superview": (_LayoutConstantAttribute, LayoutAttribute.trailing, LayoutAttribute.trailing),
        "align_top_with_superview": (_LayoutConstantAttribute, LayoutAttribute.top, LayoutAttribute.top),
        "align_bottom_with_superview": (_LayoutConstantAttribute, LayoutAttribute.bottom, LayoutAttribute.bottom),
        "align_left_with_superview": (_LayoutConstantAttribute, LayoutAttribute.left, LayoutAttribute.left),
        "align_right_with_superview": (_LayoutConstantAttribute, LayoutAttribute.right, LayoutAttribute.right),

        "relative_superview_width": (_LayoutMultiplierAttribute, LayoutAttribute.width, LayoutAttribute.width),
        "relative_superview_height": (_LayoutMultiplierAttribute, LayoutAttribute.height, LayoutAttribute.height),

        "relative_width_to": (_LayoutMultiplierAttribute, LayoutAttribute.width, LayoutAttribute.width),
        "relative_height_to": (_LayoutMultiplierAttribute, LayoutAttribute.height, LayoutAttribute.height),

        "left_offset_to": (_LayoutConstantAttribute, LayoutAttribute.left, LayoutAttribute.right),
        "right_offset_to": (_LayoutConstantAttribute, LayoutAttribute.right, LayoutAttribute.left),
        "top_offset_to": (_LayoutConstantAttribute, LayoutAttribute.top, LayoutAttribute.bottom),
        "bottom_offset_to": (_LayoutConstantAttribute, LayoutAttribute.bottom, LayoutAttribute.top),
        "leading_offset_to": (_LayoutConstantAttribute, LayoutAttribute.leading, LayoutAttribute.trailing),
        "trailing_offset_to": (_LayoutConstantAttribute, LayoutAttribute.trailing, LayoutAttribute.leading)
    }

    def __init__(self, view):
        assert(isinstance(view, LayoutProxy))

        self._view = view
        self._attributes = {}

    def _create_attribute(self, cls, attribute, other_attribute, other):
        assert(isinstance(attribute, LayoutAttribute))
        if other:
            assert(isinstance(other_attribute, LayoutAttribute))
            assert(isinstance(other, ui.View))

        name = '{}{}{}'.format(int(attribute), id(other), int(other_attribute))

        layout_attribute = self._attributes.get(name, None)
        if layout_attribute:
            return layout_attribute

        layout_attribute = cls(self._view, attribute, other, other_attribute)
        self._attributes[name] = layout_attribute
        return layout_attribute

    def _attribute(self, name, definition):
        cls = definition[0]
        attribute = definition[1]
        other_attribute = definition[2]

        assert(isinstance(attribute, LayoutAttribute))
        assert(isinstance(other_attribute, LayoutAttribute))

        if 'superview' in name:
            return self._create_attribute(cls, attribute, other_attribute, self._view.superview)
        elif name.endswith('_to'):
            return partial(self._create_attribute, cls, attribute, other_attribute)
        else:
            assert(other_attribute is LayoutAttribute.notAnAttribute)
            return self._create_attribute(cls, attribute, other_attribute, None)

    def __getattr__(self, name):
        if name in self._definitions:
            return self._attribute(name, self._definitions[name])

        return super().__getattr__(name)

    def center_in_superview(self):
        self.align_center_x_in_superview.equal = 0
        self.align_center_y_in_superview.equal = 0


class LayoutProxy(ui.View):
    _attributes = [
        '_view', '_view_objc', '_layout', '_objc', 'layout', 'view'
    ]

    def __init__(self, view):
        assert(isinstance(view, ui.View))

        self._view = view
        self.add_subview(self._view)

        self._view_objc = ObjCInstance(self._view)
        self._objc = ObjCInstance(self)

        self._view_objc.setTranslatesAutoresizingMaskIntoConstraints_(False)
        self._objc.setTranslatesAutoresizingMaskIntoConstraints_(False)

        attributes = [LayoutAttribute.left, LayoutAttribute.right, LayoutAttribute.top, LayoutAttribute.bottom]
        for attribute in attributes:
            constraint = _LayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
                self._view_objc, int(attribute), int(LayoutRelation.equal), self._objc, int(attribute), 1.0, 0
            )
            self._objc.addConstraint_(constraint)

        self._layout = None

    @property
    def layout(self):
        if not self._layout:
            self._layout = Layout(self)

        return self._layout

    @property
    def view(self):
        return self._view

    def __getattr__(self, name):
        if name in self._attributes:
            return super().__getattr__(name)
        return self._view.__getattr__(name)

    def __setattr__(self, name, value):
        if name in self._attributes:
            return super().__setattr__(name, value)
        return self._view.__setattr__(name, value)



#!python3

import blackmamba.ide.script as script


class ActionInfo(object):
    def __init__(self, action_info):
        self.script_name = str(action_info['scriptName'])
        self.icon_name = str(action_info['iconName'])

        if action_info['title']:
            self.title = str(action_info['title'])
        else:
            self.title = None

        if action_info['iconColor']:
            self.icon_color = '#{}'.format(action_info['iconColor'])
        else:
            self.icon_color = '#FFFFFF'

    def run(self, delay=None):
        script.run_script(self.script_name, delay=delay)


def get_actions():
    from objc_util import ObjCClass
    defaults = ObjCClass('NSUserDefaults').standardUserDefaults()
    return [ActionInfo(ai) for ai in defaults.objectForKey_('EditorActionInfos')]


def get_action(title):
    for action in get_actions():
        if action.title == title:
            return action
    return None


def run_action(title, delay=None):
    action = get_action(title)

    if not action:
        return

    action.run(delay)



#!python3
"""
Script name ``outline_quickly.py``.

Shows source code outline. You can filter functions, ... by name. Use arrows key to
change selection and then hit ``Enter`` to scroll to the symbol.
"""

import editor
import ast
import os
from enum import Enum
from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
from ui import Image
import blackmamba.ide.source as source


class OutlineNodeItem(PickerItem):
    class Style(Enum):
        cls = 'class'
        fn = 'function'

    def __init__(self, style, name, line, column, level, breadcrumb):
        super().__init__('{}{}'.format(level * '    ', name),
                         '{} (line {})'.format(breadcrumb, line))
        self.style = style
        self.name = name
        self.line = line
        self.column = column
        self.level = level
        self.breadcrumb = breadcrumb
        self.image = Image(style.value)


class OutlineDataSource(PickerDataSource):
    def __init__(self, text, filename):
        super().__init__()

        r = ast.parse(text)
        self.items = OutlineDataSource._generate_nodes(r, breadcrumb=filename)

    @staticmethod
    def _generate_nodes(parent, level=0, breadcrumb=''):
        nodes = []

        for child in parent.body:
            if isinstance(child, ast.ClassDef):
                style = OutlineNodeItem.Style.cls
            elif isinstance(child, ast.FunctionDef):
                style = OutlineNodeItem.Style.fn
            else:
                style = None

            if style:
                node = OutlineNodeItem(style, child.name, child.lineno, child.col_offset, level, breadcrumb)
                nodes.append(node)
                if breadcrumb:
                    bc = '{} • {}'.format(breadcrumb, child.name)
                else:
                    bc = child.name
                child_nodes = OutlineDataSource._generate_nodes(child, level + 1, bc)
                if child_nodes:
                    nodes.extend(child_nodes)

        return nodes


def main():
    filename = editor.get_path()
    if not filename:
        return

    if not filename.endswith('.py'):
        return

    text = editor.get_text()
    if not text:
        return

    def scroll_to_node(node, shift_enter):
        source.scroll_to_line(node.line)

    v = PickerView()
    v.name = 'Outline'
    v.datasource = OutlineDataSource(text, os.path.basename(filename))
    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - scroll to location'
        '\n'
        'Esc - close • Cmd . - close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter nodes...'
    v.did_select_item_action = scroll_to_node
    v.present('sheet')
    v.wait_modal()


if __name__ == '__main__':
    main()



#!python3
"""
Script name ``analyze.py``.

Source code analysis with ``pep8`` and ``pyflakes``. Results are displayed as
Pythonista annotations. If there's no error / warning, HUD informs you about that.

This script is configurable, see :ref:`configuration`.
"""

import io
import pep8
import re
from enum import Enum
import pyflakes.api as pyflakes
import editor
import console
from blackmamba.ide.annotation import Annotation, Style
from itertools import groupby
from blackmamba.config import get_config_value


def _hud_alert_delay():
    return get_config_value('analyzer.hud_alert_delay', 1.0)


def _ignore_codes():
    return get_config_value('analyzer.ignore_codes', ['W391', 'W293'])


def _max_line_length():
    return get_config_value('analyzer.max_line_length', 79)


def _remove_whitespaces():
    return get_config_value('analyzer.remove_whitespaces', True)


_REMOVE_TRAILING_WHITESPACES_REGEX = re.compile('[ \t]+$', re.MULTILINE)
_REMOVE_TRAILING_LINES_REGEX = re.compile('\s+\Z', re.MULTILINE)

#
# Common for pep8 & pyflakes
#


class _Source(Enum):
    pep8 = 'PEP8'
    pyflakes = 'pyflakes'


class _AnalyzerAnnotation(Annotation):
    def __init__(self, line, text, source, style):
        super().__init__(line, text, style)
        self.source = source

    def __lt__(self, other):
        if self.source is _Source.pep8 and other.source is _Source.pyflakes:
            return True

        if self.style is Style.warning and other.style is Style.error:
            return True

        return False


#
# pep8
#

class _Pep8AnnotationReport(pep8.BaseReport):
    def __init__(self, options):
        super().__init__(options)
        self.annotations = []

    def error(self, line_number, offset, text, check):
        # If super doesn't return code, this one is ignored
        if not super().error(line_number, offset, text, check):
            return

        annotation = _AnalyzerAnnotation(self.line_offset + line_number, text, _Source.pep8, Style.warning)
        self.annotations.append(annotation)


def _pep8_annotations(text, ignore=None, max_line_length=None):
    # pep8 requires you to include \n at the end of lines
    lines = text.splitlines(True)

    style_guide = pep8.StyleGuide(reporter=_Pep8AnnotationReport, )
    options = style_guide.options

    if ignore:
        options.ignore = tuple(ignore)
    else:
        options.ignore = tuple()

    if max_line_length:
        options.max_line_length = max_line_length

    checker = pep8.Checker(None, lines, options, None)
    checker.check_all()

    return checker.report.annotations


#
# pyflakes
#

_LINE_COL_MESSAGE_REGEX = re.compile('^(\d+):(\d+): (.*)$')
_LINE_MESSAGE_REGEX = re.compile('^(\d+): (.*)$')


def _get_annotations(path, stream, style):
    l = len(path)

    annotations = []
    for line in stream.getvalue().splitlines():
        if not line.startswith(path):
            continue

        line = line[(l + 1):]  # Strip 'filename:'
        match = _LINE_COL_MESSAGE_REGEX.fullmatch(line)

        if not match:
            match = _LINE_MESSAGE_REGEX.fullmatch(line)

        if not match:
            continue

        line = int(match.group(1))

        if match.lastindex == 3:
            annotation = _AnalyzerAnnotation(
                line, 'Col {}: {}'.format(match.group(2), match.group(3)),
                _Source.pyflakes, style
            )
        else:
            annotation = _AnalyzerAnnotation(
                line, match.group(2),
                _Source.pyflakes, style
            )

        annotations.append(annotation)

    return annotations


def _pyflakes_annotations(path, text):
    warning_stream = io.StringIO()
    error_stream = io.StringIO()
    reporter = pyflakes.modReporter.Reporter(warning_stream, error_stream)

    pyflakes.check(text, path, reporter)

    warnings = _get_annotations(path, warning_stream, Style.warning)
    errors = _get_annotations(path, error_stream, Style.error)

    return warnings + errors


def _annotate(line, annotations, scroll):
    by_priority = sorted(annotations, reverse=True)

    style = by_priority[0].style.value
    text = ',\n'.join([a.text for a in by_priority])

    editor.annotate_line(line, text, style, True, scroll=scroll)


def _remove_trailing_whitespaces(text):
    return _REMOVE_TRAILING_WHITESPACES_REGEX.sub('', text)


def _remove_trailing_lines(text):
    return _REMOVE_TRAILING_LINES_REGEX.sub('', text)


def _editor_text():
    text = editor.get_text()

    range_end = len(text)

    if _remove_whitespaces():
        text = _remove_trailing_whitespaces(text)
        text = _remove_trailing_lines(text)
        editor.replace_text(0, range_end, text)
        # Pythonista is adding '\n' automatically, so, if we removed them
        # all we have to simulate Pythonista behavior by adding '\n'
        # for pyflakes & pep8 analysis
        return text + '\n'

    return text


def main():
    path = editor.get_path()

    if not path:
        return

    if not path.endswith('.py'):
        return

    editor.clear_annotations()

    text = _editor_text()

    annotations = _pep8_annotations(
        text,
        ignore=_ignore_codes(),
        max_line_length=_max_line_length()
    )

    annotations += _pyflakes_annotations(path, text)

    if not annotations:
        console.hud_alert('No Issues Found', 'iob:checkmark_32', _hud_alert_delay())
        return None

    scroll = True
    by_line = sorted(annotations, key=lambda x: x.line)
    for l, a in groupby(by_line, lambda x: x.line):
        _annotate(l, a, scroll)
        scroll = False


if __name__ == '__main__':
    main()



#!python3
"""
I strongly suggests to check
`config.py:_DEFAULTS <https://github.com/zrzka/blackmamba/blob/master/blackmamba/config.py>`_
from time to time. Just to check if this documentation isn't outdated.

Sections
========

General
-------

Defaults:

.. code-block:: python

    'general': {
        'jedi': False,
        'register_key_commands': True,
        'page_line_count': 40
    }


``jedi: bool`` - `JEDI <http://jedi.readthedocs.io/en/latest/>`_ is used as a backend for Find usages,
Jump to definition and Show documentation :ref:`scripts`. But because JEDI is also used by the Pythonista,
JEDI is not thread safe and I don't know when and how it is used, it's disabled by default. You have to
set it to ``True`` to enable mentioned scripts.

``register_key_commands: bool`` - set to ``False`` to disable default :ref:`shortcuts`.

``page_line_count: int`` - number of lines to scroll up / down for page up / down.


Updates
-------

Defaults:

.. code-block:: python

    'update': {
        'enabled': True,
        'interval': 3600
    }

``enabled: bool`` - set to ``False`` to disable check for updates.

``interval: int`` - check for updates time interval (in seconds).



File picker
-----------

Defaults:

.. code-block:: python

    'file_picker': {
        'ignore_folders': {
            '': ['.git'],
            '.': ['.Trash', 'Examples',
                  'site-packages', 'site-packages-2', 'site-packages-3']
        }
    }

``ignore_folders: dict`` - key is a parent directory (not full path, just name) and value is a list of
folders to ignore. You can use two special values as keys:


* ``''`` - any parent directory,
* ``'.'`` - parent directory is ``~/Documents``.

Default value says that ``.git`` folder inside any folder is ignored. ``.Trash``,
``Examples``, ... folders inside ``~/Documents`` folder are ignored as well.

Affects Open quickly and Run quickly scripts, see :ref:`scripts`.


Analyzer
--------

Defaults:

.. code-block:: python

    'analyzer': {
        'hud_alert_delay': 1.0,
        'ignore_codes': None,
        'max_line_length': 127,
        'remove_whitespaces': True
    }

Affects Analyze script, see :ref:`scripts`.


Tester
------

Defaults:

.. code-block:: python

    'tester': {
        'hud_alert_delay': 1.0,
        'hide_console': True
    }

Affects Run unit tests script, see :ref:`scripts`.


Drag and Drop
-------------

Defaults:

.. code-block:: python

    'drag_and_drop': {
        'ignore_folders': {
            '': ['.git'],
            '.': ['.Trash', 'Examples',
                  'site-packages', 'site-packages-2', 'site-packages-3', 'stash_extensions']
        }
    }

Affects Drag & Drop script, see :ref:`scripts`.


Sample
======

See `pythonista_startup.py <https://github.com/zrzka/blackmamba/blob/master/pythonista_startup.py>`_
for more examples.
"""

from collections import Mapping
from copy import deepcopy

__all__ = ['get_config_value']

_DEFAULTS = {
    'general': {
        'jedi': False,
        'register_key_commands': True,
        'page_line_count': 40
    },
    'update': {
        'enabled': True,
        'interval': 3600
    },
    'file_picker': {
        'ignore_folders': {
            '': ['.git'],
            '.': ['.Trash', 'Examples',
                  'site-packages', 'site-packages-2', 'site-packages-3']
        }
    },
    'analyzer': {
        'hud_alert_delay': 1.0,
        'ignore_codes': None,
        'max_line_length': 127,
        'remove_whitespaces': True
    },
    'tester': {
        'hud_alert_delay': 1.0,
        'hide_console': True
    },
    'drag_and_drop': {
        'ignore_folders': {
            '': ['.git'],
            '.': ['.Trash', 'Examples',
                  'site-packages', 'site-packages-2', 'site-packages-3', 'stash_extensions']
        }
    }
}

_config = dict(_DEFAULTS)


def get_config_value(key_path, default=None):
    v = _config
    try:
        for k in key_path.split('.'):
            v = v[k]
        return v
    except TypeError:
        return default
    except KeyError:
        return default


def _update(d, u):
    for k, v in u.items():
        if isinstance(d, Mapping):
            if isinstance(v, Mapping):
                d[k] = _update(d.get(k, {}), v)
            else:
                d[k] = v
        else:
            d = {k: v}
    return d


def update_config_with_dict(d):
    global _config
    _config = _update(deepcopy(_config), d)



#!python3
"""
Script name ``run_unit_tests.py``.

Runs unit tests and show results as Pythonista annotations.

This script is configurable, see :ref:`configuration`.
"""

import pytest
import editor
import os
import xml.etree.ElementTree as ET
import console
from blackmamba.ide.annotation import Annotation, Style
from blackmamba.config import get_config_value
from blackmamba.ide.tab import save
import re

_LOG_FILE_PATH = os.path.expanduser('~/Documents/.blackmamba_pytest_log.xml')


def _hud_alert_delay():
    return get_config_value('tester.hud_alert_delay', 1.0)


def _hide_console():
    return get_config_value('tester.hide_console', True)


def _remove_log_file():
    if os.path.isfile(_LOG_FILE_PATH):
        os.remove(_LOG_FILE_PATH)


def _parse_line(file, text, line):
    # Test case line number points to the test function, not actual issue
    # If we can parse, we'll return the correct one, otherwise fallback
    # to the test function

    pattern = re.escape(file) + '\:(\d+)\:'

    match = re.search(pattern, text, re.MULTILINE)
    if match:
        return int(match.group(1))
    return line


def _parse_log_file():
    if not os.path.isfile(_LOG_FILE_PATH):
        return

    try:
        suite = ET.parse(_LOG_FILE_PATH).getroot()
    except:
        return

    if not suite.tag == 'testsuite':
        return

    result = []
    for tc in list(suite):
        if not tc.tag == 'testcase':
            continue

        for failure in list(tc):
            if not failure.tag == 'failure':
                continue

            file = tc.get('file')
            text = failure.get('message')
            line = _parse_line(file, failure.text, int(tc.get('line')))

            annotation = Annotation(line, text, Style.error)  # , os.path.basename(file))
            result.append(annotation)

    return (suite.attrib, result)


def _show_results(attrib, all_passed=True):
    x = {
        'errors': 'errored',
        'failures': 'failed',
        'skips': 'skipped'
    }

    tests = int(attrib.get('tests', 0))
    time = attrib.get('time', '?')

    messages = []

    for k, v in x.items():
        c = int(attrib.get(k, 0))

        if c > 0:
            tests -= c
            messages.append('{} {}'.format(c, v))

    messages.append('{} passed'.format(tests))

    console.hud_alert(
        ', '.join(messages) + ' in {} seconds'.format(time),
        'success' if all_passed else 'error',
        _hud_alert_delay())


def _run_unit_tests(path):
    editor.clear_annotations()

    _remove_log_file()
    pytest.main(['-q', '--junitxml={}'.format(_LOG_FILE_PATH), path])
    attrib, annotations = _parse_log_file()

    _show_results(attrib, not annotations)

    scroll = True
    for a in annotations:
        editor.annotate_line(a.line, a.text, a.editor_annotation_style, True, filename=a.filename, scroll=scroll)
        scroll = False

    if _hide_console():
        console.hide_output()


def main():
    path = editor.get_path()

    if not path:
        return

    if not path.endswith('.py'):
        return

    save(all=True)
    _run_unit_tests(path)


if __name__ == '__main__':
    main()



#!python3

import ui
from blackmamba.uikit.table import UITableViewCellStyle
from blackmamba.uikit.keyboard import (
    UIKeyModifier, UIEventKeyCode,
    register_key_event_handler, unregister_key_event_handlers,
    is_in_hardware_keyboard_mode
)
from blackmamba.uikit.autolayout import LayoutProxy


class PickerItem:
    def __init__(self, title, subtitle=None, image=None):
        self._title = title
        self._normalized_title = title.lower()

        self.subtitle = subtitle
        self.image = image
        self.title = title

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, new_value):
        self._title = new_value
        self._normalized_title = new_value.lower() if new_value else None

    @property
    def sort_value(self):
        return self._normalized_title

    @property
    def match_value(self):
        return self._normalized_title

    def matches(self, search_terms):
        if not search_terms:
            return True

        start = 0
        for t in search_terms:
            start = self.match_value.find(t, start)

            if start == -1:
                return False

            start += len(t)

        return True

    def __lt__(self, other):
        return self.sort_value < other.sort_value

    def __str__(self):
        return self.title


class PickerDataSource(object):
    def __init__(self):
        self.tableview = None
        self.action = None
        self._filter_by = None
        self._selected_row = -1
        self._items = None
        self._filtered_items = []
        self._filter = None

    def _filter_items(self):
        if self._filter:
            search_terms = [
                x.strip()
                for x in self._filter.split(' ')
                if x.strip()
            ]

            self._filtered_items = [
                i for i in self._items if i.matches(search_terms)
            ]
        else:
            self._filtered_items = self._items
        self._selected_row = -1
        self.reload()

    def filter_by(self, filter):
        self._filter = filter
        self._filter_items()

    @property
    def filtered_items(self):
        return self._filtered_items

    @property
    def items(self):
        return self._items

    @items.setter
    def items(self, value):
        self._items = value
        self._filter_items()

    @property
    def selected_row(self):
        return self._selected_row

    @selected_row.setter
    def selected_row(self, value):
        if self.tableview:
            self.tableview.selected_row = (0, value)
        self._selected_row = value

    @property
    def selected_item(self):
        if self.selected_row >= 0:
            return self._filtered_items[self.selected_row]
        return None

    def select_next_item(self):
        count = len(self._filtered_items)

        if not count:
            return

        row = (self.selected_row + 1) % count
        self.selected_row = row

    def select_previous_item(self):
        count = len(self._filtered_items)

        if not count:
            return

        row = self.selected_row - 1
        if row < 0:
            row = count - 1

        self.selected_row = row

    def reload(self):
        if self.tableview:
            self.tableview.reload()

    def tableview_number_of_sections(self, tv):
        self.tableview = tv
        return 1

    def tableview_number_of_rows(self, tv, section):
        return len(self._filtered_items)

    def tableview_can_delete(self, tv, section, row):
        return False

    def tableview_can_move(self, tv, section, row):
        return False

    def tableview_did_select(self, tv, section, row):
        self._selected_row = row
        if self.action and row >= 0:
            self.action(self)

    def tableview_cell_for_row(self, tv, section, row):
        item = self._filtered_items[row]
        cell = ui.TableViewCell(UITableViewCellStyle.subtitle.value)
        cell.text_label.number_of_lines = 1
        cell.text_label.text = item.title
        cell.detail_text_label.text = item.subtitle
        cell.detail_text_label.text_color = (0, 0, 0, 0.5)
        if item.image:
            cell.image_view.image = item.image
        return cell


class PickerView(ui.View):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if 'background_color' not in kwargs:
            self.background_color = 'white'

        if 'frame' not in kwargs:
            self.width = min(ui.get_window_size()[0] * 0.8, 700)
            self.height = ui.get_window_size()[1] * 0.8

        self._tableview = ui.TableView()
        self._textfield = ui.TextField()
        self._help_label = ui.Label()
        self._datasource = None
        self._handlers = None
        self.shift_enter_enabled = True
        self.did_select_item_action = None

        tf = LayoutProxy(self._textfield)
        self.add_subview(tf)
        tf.layout.align_left_with_superview.equal = 8
        tf.layout.align_top_with_superview.equal = 8
        tf.layout.align_right_with_superview.equal = -8
        tf.layout.height.equal = 31
        tf.delegate = self

        tv = LayoutProxy(self._tableview)
        self.add_subview(tv)
        tv.layout.align_left_to(tf).equal = 0
        tv.layout.align_right_to(tf).equal = 0
        tv.layout.top_offset_to(tf).equal = 8
        tv.allows_selection = True
        tv.allows_multiple_selection = False

        hl = LayoutProxy(self._help_label)
        self.add_subview(hl)
        hl.layout.align_left_to(tv).equal = 0
        hl.layout.align_right_to(tv).equal = 0
        hl.layout.top_offset_to(tv).equal = 8
        hl.layout.align_bottom_with_superview.equal = -8
        hl.layout.height.max = 66
        hl.font = ('<system>', 13.0)
        hl.alignment = ui.ALIGN_CENTER
        hl.text_color = (0, 0, 0, 0.5)
        hl.number_of_lines = 2

        if is_in_hardware_keyboard_mode:
            tf.view.begin_editing()

        self._register_key_event_handlers()

    def _register_key_event_handlers(self):
        def handle_key_up():
            if not self._datasource:
                return

            self._datasource.select_previous_item()

        def handle_key_down():
            if not self._datasource:
                return

            self._datasource.select_next_item()

        def handle_shift_enter():
            if self.shift_enter_enabled:
                self._did_select_item(True)

        def handle_enter():
            self._did_select_item()

        def handle_escape():
            self.close()

        self._handlers = [
            register_key_event_handler(UIEventKeyCode.up, handle_key_up),
            register_key_event_handler(UIEventKeyCode.down, handle_key_down),
            register_key_event_handler(UIEventKeyCode.enter, handle_enter),
            register_key_event_handler(UIEventKeyCode.enter, handle_shift_enter,
                                       modifier=UIKeyModifier.shift),
            register_key_event_handler(UIEventKeyCode.escape, handle_escape),
            register_key_event_handler(UIEventKeyCode.dot, handle_escape,
                                       modifier=UIKeyModifier.command)
        ]

    def will_close(self):
        if self._handlers:
            unregister_key_event_handlers(self._handlers)

    @property
    def datasource(self):
        return self._datasource

    @datasource.setter
    def datasource(self, value):
        self._datasource = value

        def did_select_item(ds):
            self._did_select_item()

        self._datasource.action = did_select_item
        self._tableview.data_source = self._datasource
        self._tableview.delegate = self._datasource

    @property
    def help_label(self):
        return self._help_label

    @property
    def textfield(self):
        return self._textfield

    def _did_select_item(self, shift_enter=False):
        if not self._datasource:
            return

        item = self._datasource.selected_item
        if not item:
            return

        self.close()
        if self.did_select_item_action:
            self.did_select_item_action(item, shift_enter)

    def textfield_should_return(self, textfield):
        return False

    def textfield_should_change(self, textfield, range, replacement):
        if self._datasource:
            string = textfield.text[:range[0]] + replacement + textfield.text[range[1]:]
            self._datasource.filter_by(string)
        return True



#!python3
"""
Keyboard shortcuts.

Key commands
============

Key command is kind of global shortcut in Pythonista. They're visible if you
hold down ``Cmd`` key for example. Black Mamba does use key commands for
shortcuts like toggle comments, etc.

Related functions:

    * :obj:`UIKeyModifier`
    * :obj:`UIKeyInput`
    * :func:`register_key_command`

Following example shows how to print *Hallo* with ``Cmd H`` keyboard shortcut.

.. code-block:: python

    from blackmamba.uikit.keyboard import (
        register_key_command, UIKeyModifier
    )

    def hallo():
        print('Hallo')

    register_key_command(
        'h',
        UIKeyModifier.command,
        hallo,
        'Print Hallo'  # Optional discoverability title (hold down Cmd)
    )

Key event handlers
==================

Key event handler is kind of local keyboard shortcut. They're not visible
if you hold down ``Cmd`` key. Designed to be used in custom dialogs.

Related functions:

    * :obj:`UIKeyModifier`
    * :obj:`UIEventKeyCode`
    * :func:`register_key_event_handler`
    * :func:`unregister_key_event_handler`

Following example shows how to close dialog with ``Cmd .`` keyboard shortcut.

.. code-block:: python

    from blackmamba.uikit.keyboard import (
        register_key_event_handler, UIEventKeyCode, UIKeyModifier,
        unregister_key_event_handlers
    )

    class MyView(ui.View):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            def close_me():
                self.close()

            self._handlers = [
                register_key_event_handler(UIEventKeyCode.dot, close_me, modifier=UIKeyModifier.command)
            ]

        def will_close(self):
            unregister_key_event_handlers(self._handlers)

Reference
=========
"""
from ctypes import CFUNCTYPE, c_void_p, c_char_p
from objc_util import retain_global, ObjCInstance, UIApplication, c, ns, on_main_thread, sel, ObjCClass
from blackmamba.util.runtime import swizzle
from blackmamba.log import error, info
import blackmamba.system as system
from enum import Enum, IntEnum, IntFlag

if system.IOS:
    _UIKeyboardImpl = ObjCClass('UIKeyboardImpl')
else:
    _UIKeyboardImpl = None


def is_in_hardware_keyboard_mode():
    """
    Check if HW keyboard is connected.

    :return: ``True`` if HW keyboard is connected otherwise ``False``
    """
    if not _UIKeyboardImpl:
        return False

    return _UIKeyboardImpl.sharedInstance().isInHardwareKeyboardMode()


UIKeyCommand = ObjCClass('UIKeyCommand')


class UIKeyModifier(IntFlag):
    """
    Key modifiers enumeration.

    Modifiers can be combined like:

    .. code-block:: python

        UIKeyModifier.command | UIKeyModifier.shift

    See also:

        * :func:`register_key_command`
        * :func:`register_key_event_handler`
    """

    none = 0
    """No modifier key"""

    alphaShift = 1 << 16  # CapsLock
    """Caps Lock key"""

    shift = 1 << 17  # Shift
    """Shift key"""

    control = 1 << 18  # Control
    """Control key"""

    alternate = 1 << 19  # Option
    """Option key"""

    command = 1 << 20  # Command
    """Command key"""

    numericPad = 1 << 21  # Key on numeric keypad
    """Key is on numberic pad"""

    @property
    def selector_name(self):
        flags = [
            name.title()
            for name, value in UIKeyModifier.__members__.items()
            if value & self is value and not value == 0
        ]
        if flags:
            return ''.join(flags)
        else:
            return ''


class UIEventType(IntEnum):
    touches = 0
    motion = 1
    remoteControl = 2
    presses = 3
    physicalKeyboard = 4


class UIEventSubtype(IntEnum):
    none = 0


class UIEventKeyCode(IntEnum):
    """
    Event key codes.

    Not all key codes are listed / included here. Feel free to create pull request with more
    key codes if you'd like to use them.

    See also:

        * :func:`register_key_event_handler`
    """

    right = 79
    """Right arrow key"""

    left = 80
    """Left arrow key"""

    down = 81
    """Down arrow key"""

    up = 82
    """Up arrow key"""

    enter = 40
    """Enter key"""

    space = 44
    """Space key"""

    backspace = 42
    """Backspace key"""

    escape = 41
    """Escape key"""

    leftSquareBracket = 47
    dot = 55
    """Dot key"""


class UIKeyInput(str, Enum):
    """
    Enumeration of special key input values.

    See also:

        * :func:`register_key_command`
    """

    leftArrow = 'UIKeyInputLeftArrow'
    """Left arrow"""

    rightArrow = 'UIKeyInputRightArrow'
    """Right arrow"""

    upArrow = 'UIKeyInputUpArrow'
    """Up arrow"""

    downArrow = 'UIKeyInputDownArrow'
    """Down arrow"""

    @property
    def selector_name(self):
        return self.name.title()


_UIKeyInputNames = {
    '/': 'Slash',
    '.': 'Dot',
    ',': 'Comma',
    '+': 'Plus',
    '-': 'Minus',
    ' ': 'Space',
    '_': 'Underscore',
    '\t': 'Tab',
    '[': 'LeftSquareBracket',
    ']': 'RightSquareBracket',
    '?': 'QuestionMark'
}

_key_commands = []


def _blackmamba_keyCommands(_self, _cmd):
    """Swizzled version of keyCommands(). It calls original method to
    get Pythonista shortcuts and then appends custom ones."""
    obj = ObjCInstance(_self)
    commands = list(obj.originalkeyCommands())
    commands.extend(_key_commands)
    return ns(commands).ptr


def _input_selector_name(input):
    if isinstance(input, UIKeyInput):
        return input.selector_name

    assert(isinstance(input, str))

    if len(input) == 1:
        input = input.upper()

        if (input >= 'A' and input <= 'Z') or (input >= '0' and input <= '9'):
            return input

    if input not in _UIKeyInputNames:
        raise ValueError('Unsupported key command input: {}'.format(input))

    return _UIKeyInputNames[input]


def _key_command_selector_name(input, modifier):
    assert(isinstance(modifier, UIKeyModifier))

    return 'blackMambaHandleKey{}{}'.format(
        modifier.selector_name,
        _input_selector_name(input)
    )


def _shortcut_name(input, modifier):
    return '{} {}'.format(
        modifier.selector_name,
        _input_selector_name(input)
    )


@system.Pythonista(appex=False)
@on_main_thread
def _register_key_command(input, modifier_flags, function, title=None):
    if not UIApplication.sharedApplication().respondsToSelector_(sel('originalkeyCommands')):
        swizzle('UIApplication', 'keyCommands', _blackmamba_keyCommands)

    selector_name = _key_command_selector_name(input, modifier_flags)
    selector = sel(selector_name)
    obj = UIApplication.sharedApplication().keyWindow()

    info('Registering key command "{}" ({})'.format(
        _shortcut_name(input, modifier_flags),
        title or 'No discoverability title'
    ))

    if not callable(function):
        error('Skipping, provided function is not callable')
        return False

    if obj.respondsToSelector_(selector):
        error('Skipping, method {} already registered'.format(selector_name))
        return False

    def key_command_action(_sel, _cmd, sender):
        function()

    IMPTYPE = CFUNCTYPE(None, c_void_p, c_void_p, c_void_p)
    imp = IMPTYPE(key_command_action)
    retain_global(imp)

    cls = c.object_getClass(obj.ptr)
    type_encoding = c_char_p('v@:@'.encode('utf-8'))
    did_add = c.class_addMethod(cls, selector, imp, type_encoding)
    if not did_add:
        error('Failed to add key command method {}'.format(selector_name))
        return False

    if title:
        kc = UIKeyCommand.keyCommandWithInput_modifierFlags_action_discoverabilityTitle_(
            ns(input), modifier_flags, selector, ns(title))
    else:
        kc = UIKeyCommand.keyCommandWithInput_modifierFlags_action_(ns(input), modifier_flags, selector)

    _key_commands.append(kc)
    return True


def register_key_command(input, modifier_flags, function, title=None):
    """
    Register key command.

    .. note:: There's no function to unregister key commands.

    :param input: ``str`` or :obj:`UIKeyInput`
    :param modifier_flags: :obj:`UIKeyModifier`
    :param function: Function to call
    :param title: Optional discoverability title
    :return: ``True`` if key command was registered otherwise ``False``
    """
    return _register_key_command(input, modifier_flags, function, title)


_key_event_handlers = []


class KeyEventHandler(object):
    def __init__(self, key_code, modifier, fn):
        if isinstance(key_code, UIEventKeyCode):
            self.key_code = key_code.value
        else:
            self.key_code = key_code
        self.modifier = modifier
        self.fn = fn


def _blackmamba_handleKeyUIEvent(_self, _cmd, event):
    e = ObjCInstance(event)

    consume = False
    if e.type() == UIEventType.physicalKeyboard.value and e.subtype() == UIEventSubtype.none.value:
        for h in _key_event_handlers:
            if h.key_code == e._keyCode() and h.modifier.value == e._modifierFlags():
                if e._isKeyDown():
                    consume = True
                else:
                    h.fn()
                break

    if not consume:
        ObjCInstance(_self).originalhandleKeyUIEvent_(e)


@on_main_thread
def _register_key_event_handler(key_code, func, *, modifier=UIKeyModifier.none):
    if not UIApplication.sharedApplication().respondsToSelector_(sel('originalhandleKeyUIEvent:')):
        swizzle('UIApplication', 'handleKeyUIEvent:', _blackmamba_handleKeyUIEvent)

    @system.catch_exceptions
    def invoke_func():
        func()

    handler = KeyEventHandler(key_code, modifier, invoke_func)
    _key_event_handlers.append(handler)
    return handler


def register_key_event_handler(key_code, func, *, modifier=UIKeyModifier.none):
    """
    Register key event handler.

    Usable in dialogs for example. Do not forget to unregister key event
    handler in ``will_close`` function of your ``ui.View``.

    :param key_code: :obj:`UIEventKeyCode` or ``int``
    :param func: Function to call
    :param modifier: :obj:`UIKeyModifier`
    :return: Handler to use in :func:`unregister_key_event_handler`
    """
    return _register_key_event_handler(key_code, func, modifier=modifier)


@on_main_thread
def _unregister_key_event_handler(handler):
    try:
        _key_event_handlers.remove(handler)
    except ValueError:
        pass


def unregister_key_event_handler(handler):
    """
    Unregister key event handler.

    It is safe to call this function multiple times with the same handler. Handler
    is silently ignored if it's not registered.

    :param handler: Handler from :func:`register_key_event_handler`
    """
    _unregister_key_event_handler(handler)


def unregister_key_event_handlers(handlers):
    """
    Unregister list of key event handlers.

    Convenience function, it just calls :func:`unregister_key_event_handler` for every handler.

    :param handlers: List of handlers
    """
    for handler in handlers:
        unregister_key_event_handler(handler)



#!python3

import os
import urllib
import webbrowser
import ui
from blackmamba.log import error


def script_exists(script_name, full_path=False):
    if not full_path:
        script_name = os.path.join(os.path.expanduser('~/Documents'), script_name)
    return os.path.exists(script_name)


def run_script(script_name, full_path=False, delay=None):
    if not full_path and script_name.startswith('/'):
        script_name = script_name[1:]

    if not script_exists(script_name, full_path):
        error('run_script: script does not exist {}'.format(script_name))
        return

    if full_path:
        docs_root = os.path.expanduser('~/Documents/')
        script_name = script_name[len(docs_root):]

    encoded_name = urllib.parse.quote_plus(script_name, safe='', encoding=None, errors=None)
    url = 'pythonista://{}?action=run'.format(encoded_name)

    if delay:
        def make_open_url(url):
            def open():
                webbrowser.open(url)
            return open

        ui.delay(make_open_url(url), 1.0)
    else:
        webbrowser.open(url)



from setuptools import setup
import re

_init_py = open('blackmamba/__init__.py').read()
__version__ = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', _init_py).group(1)
__author__ = re.search(r'__author__\s*=\s*[\'"]([^\'"]*)[\'"]', _init_py).group(1)

setup(
    name='blackmamba',
    version=__version__,

    description='Pythonista on steroids',
    long_description=open('README.rst', 'rt').read(),

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: iOS',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development'
    ],

    url='https://github.com/zrzka/blackmamba',
    license='MIT',

    author=__author__,

    packages=[
        'blackmamba',
        'blackmamba.uikit',
        'blackmamba.ide',
        'blackmamba.script',
        'blackmamba.util'
    ],
    package_dir={'': '.'},

    keywords=['pythonista ios']
)



#!python3
"""
Script name ``close_all_tabs_except_current_one.py``.

Closes all tabs except current one.
"""

import blackmamba.ide.tab as tab


def main():
    tab.close_tabs_except_current()


if __name__ == '__main__':
    main()



#!python3
"""
Logging module.

Why custom module instead of :mod:`logging`? Several reasons:

* not to interfere with Pythonista logging settings,
* unable to convince Pythonista to use my colors,
* etc.

Default log level is ``INFO``. You can use :func:`blackmamba.log.set_level`
to change effective log level. Available log levels are:

* ``ERROR``
* ``WARNING``
* ``INFO``
* ``DEBUG``
* ``NOTSET``

If you'd like to silent Black Mamba messages, it's recommended to set log
level to ``ERROR``.

.. code-block:: python

    import blackmamba.log as log

    log.set_level(log.ERROR)

.. warning:: This module must not introduce dependency on any other Black Mamba
    modules and must be importable on any other platform.

Reference
=========
"""

try:
    import console
except ImportError:
    console = None

ERROR = 40
WARNING = 30
INFO = 20
DEBUG = 10
NOTSET = 0

_level = INFO

_COLORS = {
    WARNING: (1, 0.5, 0),
    ERROR: (1, 0, 0)
}


def get_level():
    """
    Return effective log level.

    :return: Effective log level
    """
    return _level


def set_level(level):
    """
    Set effective log level.

    :param level: Level to set
    """
    global _level
    _level = level


def _log(level, *args, **kwargs):
    if _level > level:
        return

    color = _COLORS.get(level, None)
    if console and color:
        console.set_color(*color)

    print(*args, **kwargs)

    if console and color:
        console.set_color()


def debug(*args, **kwargs):
    """
    Log message with DEBUG level.

    :param args: Passed to :func:`print`
    :param kwargs: Passed to :func:`print`
    """
    _log(DEBUG, *args, **kwargs)


def info(*args, **kwargs):
    """
    Log message with INFO level.

    :param args: Passed to :func:`print`
    :param kwargs: Passed to :func:`print`
    """
    _log(INFO, *args, **kwargs)


def warn(*args, **kwargs):
    """
    Log message with WARNING level.

    :param args: Passed to :func:`print`
    :param kwargs: Passed to :func:`print`
    """
    _log(WARNING, *args, **kwargs)


def error(*args, **kwargs):
    """
    Log message with ERROR level.

    :param args: Passed to :func:`print`
    :param kwargs: Passed to :func:`print`
    """
    _log(ERROR, *args, **kwargs)



#!python3
"""
Script name ``action_quickly.py``.

Shows dialog with user defined action items (wrench icon). You can filter them
by title, use arrow keys to change selection and run any of them with ``Enter`` key.

Only **user defined action items** are listed.
"""

import os
from ui import TableViewCell
from blackmamba.uikit.picker import PickerView, PickerItem, PickerDataSource
from blackmamba.uikit.table import UITableViewCellStyle
from blackmamba.ide.action import ActionInfo, get_actions


class ActionPickerItem(PickerItem):
    def __init__(self, action_info):
        if not isinstance(action_info, ActionInfo):
            action_info = ActionInfo(action_info)

        if action_info.title:
            title = action_info.title
        else:
            _, tail = os.path.split(action_info.script_name)
            title, _ = os.path.splitext(tail)

        subtitle = ' • '.join(action_info.script_name.split(os.sep))

        super().__init__(title, subtitle)

        self.action_info = action_info


class ActionPickerDataSource(PickerDataSource):
    def __init__(self):
        super().__init__()
        self.items = sorted([ActionPickerItem(ai) for ai in get_actions()])

    def tableview_cell_for_row(self, tv, section, row):
        item = self.filtered_items[row]
        cell = TableViewCell(UITableViewCellStyle.subtitle.value)
        cell.text_label.number_of_lines = 1
        cell.text_label.text = item.title
        cell.detail_text_label.text = item.subtitle
        cell.detail_text_label.text_color = (0, 0, 0, 0.5)
        return cell


def main():
    def run_wrench_item(item, shift_enter):
        item.action_info.run(delay=1.0)

    v = PickerView()
    v.name = 'Action Quickly...'
    v.datasource = ActionPickerDataSource()
    v.shift_enter_enabled = False
    v.help_label.text = (
        '⇅ - select • Enter - run action item'
        '\n'
        'Esc - close • Cmd . - Close with Apple smart keyboard'
    )
    v.textfield.placeholder = 'Start typing to filter wrench items...'
    v.did_select_item_action = run_wrench_item
    v.present('sheet')
    v.wait_modal()


if __name__ == '__main__':
    main()



#!python3

from objc_util import on_main_thread, c, parse_types, ObjCClass, sel, retain_global
from ctypes import CFUNCTYPE, c_void_p, c_char_p, c_ulong, c_int, Structure, cast, POINTER
from blackmamba.log import error

SWIZZLED_SELECTOR_PREFIX = 'original'


method_exchangeImplementations = c.method_exchangeImplementations
method_exchangeImplementations.argtypes = [c_void_p, c_void_p]
method_exchangeImplementations.restype = c_void_p


@on_main_thread
def add_method(cls_name, selector_name, fn, type_encoding):
    cls = ObjCClass(cls_name).ptr

    selector = sel(selector_name)

    if c.class_getInstanceMethod(cls, selector):
        error('Failed to add method, class {} already provides method {}'.format(cls_name, selector_name))
        return

    parsed_types = parse_types(type_encoding)
    restype = parsed_types[0]
    argtypes = parsed_types[1]

    IMPTYPE = CFUNCTYPE(restype, *argtypes)
    imp = IMPTYPE(fn)
    retain_global(imp)

    did_add = c.class_addMethod(cls, selector, imp, c_char_p(type_encoding.encode('utf-8')))
    if not did_add:
        error('Failed to add class method')

    return did_add


@on_main_thread
def swizzle(cls_name, selector_name, fn):
    cls = ObjCClass(cls_name).ptr

    new_selector_name = SWIZZLED_SELECTOR_PREFIX + selector_name
    new_selector = sel(new_selector_name)

    if c.class_getInstanceMethod(cls, new_selector):
        error('Skipping swizzling, already responds to {} selector'.format(new_selector_name))
        return

    selector = sel(selector_name)
    method = c.class_getInstanceMethod(cls, selector)
    if not method:
        error('Failed to get {} instance method'.format(selector_name))
        return

    type_encoding = c.method_getTypeEncoding(method)
    parsed_types = parse_types(type_encoding)
    restype = parsed_types[0]
    argtypes = parsed_types[1]

    IMPTYPE = CFUNCTYPE(restype, *argtypes)
    imp = IMPTYPE(fn)
    retain_global(imp)

    did_add = c.class_addMethod(cls, new_selector, imp, type_encoding)

    if not did_add:
        error('Failed to add {} method'.format(new_selector_name))
        return

    new_method = c.class_getInstanceMethod(cls, new_selector)
    method_exchangeImplementations(method, new_method)


#
# https://clang.llvm.org/docs/Block-ABI-Apple.html
#

_BLOCK_HAS_COPY_DISPOSE = 1 << 25
_BLOCK_HAS_CTOR = 1 << 26
_BLOCK_IS_GLOBAL = 1 << 28
_BLOCK_HAS_STRET = 1 << 29
_BLOCK_HAS_SIGNATURE = 1 << 30


class _BlockDescriptor(Structure):
    _fields_ = [('reserved', c_ulong),
                ('size', c_ulong),
                ('copy_helper', c_void_p),     # flags & (1<<25)
                ('dispose_helper', c_void_p),  # flags & (1<<25)
                ('signature', c_char_p)]       # flags & (1<<30)


class _Block(Structure):
    _fields_ = [('isa', c_void_p),
                ('flags', c_int),
                ('reserved', c_int),
                ('invoke', c_void_p),
                ('descriptor', _BlockDescriptor)]


class ObjCBlockPointer():
    def __init__(self, block_ptr, restype=None, argtypes=None):
        self._block_ptr = block_ptr
        self._block = cast(self._block_ptr, POINTER(_Block))

        if not argtypes:
            argtypes = []

        if self._regular_calling_convention():
            # First arg is pointer to block, hide it from user
            argtypes.insert(0, c_void_p)

        if self._has_signature():
            # TODO - Validate restype & argtypes against signature
            #      - Signature is not always populated
            pass

        self._func = None

        if self._regular_calling_convention():
            IMPTYPE = CFUNCTYPE(restype, *argtypes)
            self._func = IMPTYPE(self._block.contents.invoke)

    def _flags(self):
        return self._block.contents.flags

    def _has_signature(self):
        has_signature = {
            (0 << 29): False,
            (1 << 29): False,
            (2 << 29): True,
            (3 << 29): True
        }

        return has_signature.get(self._flags() & (3 << 29), False)

    def _signature(self):
        if not self._has_signature():
            return None

        flags = self._flags()

        if flags & _BLOCK_HAS_COPY_DISPOSE:
            signature = self._block.contents.descriptor.signature
        else:
            signature = self._block.contents.descriptor.copy_helper

        return signature

    def _regular_calling_convention(self):
        flags = self._flags() & (3 << 29)
        return flags == (2 << 29)

    def _stret_calling_convention(self):
        flags = self._flags() & (3 << 29)
        return flags == (3 << 29)

    def __call__(self, *args):
        if self._regular_calling_convention():
            # First arg is pointer to block, hide it from user
            return self._func(self._block_ptr, *args)

        error('Not implemented calling convention, block not called')
